from __future__ import annotations

import typing
from typing import Any, NoReturn, Optional, Type

import numpy

if typing.TYPE_CHECKING:
    from policyengine_core.enums import Enum


class EnumArray(numpy.ndarray):
    """
    Numpy array subclass representing an array of enum items.

    EnumArrays are encoded as ``int`` arrays to improve performance
    """

    # Subclassing ndarray is a little tricky.
    # To read more about the two following methods, see:
    # https://docs.scipy.org/doc/numpy-1.13.0/user/basics.subclassing.html#slightly-more-realistic-example-attribute-added-to-existing-array.
    def __new__(
        cls,
        input_array: numpy.int_,
        possible_values: Optional[Type[Enum]] = None,
    ) -> EnumArray:
        obj = numpy.asarray(input_array).view(cls)
        obj.possible_values = possible_values
        return obj

    # See previous comment
    def __array_finalize__(self, obj: Optional[numpy.int_]) -> None:
        if obj is None:
            return

        self.possible_values = getattr(obj, "possible_values", None)

    def __eq__(self, other: Any) -> bool:
        # When comparing to an item of self.possible_values, use the item index
        # to speed up the comparison.
        if other.__class__.__name__ is self.possible_values.__name__:
            # Use view(ndarray) so that the result is a classic ndarray, not an
            # EnumArray.
            return self.view(numpy.ndarray) == other.index

        return self.view(numpy.ndarray) == other

    def __ne__(self, other: Any) -> bool:
        return numpy.logical_not(self == other)

    def _forbidden_operation(self, other: Any) -> NoReturn:
        raise TypeError(
            "Forbidden operation. The only operations allowed on EnumArrays "
            "are '==' and '!='.",
        )

    __add__ = _forbidden_operation
    __mul__ = _forbidden_operation
    __lt__ = _forbidden_operation
    __le__ = _forbidden_operation
    __gt__ = _forbidden_operation
    __ge__ = _forbidden_operation
    __and__ = _forbidden_operation
    __or__ = _forbidden_operation

    def decode(self) -> numpy.object_:
        """
        Return the array of enum items corresponding to self.

        For instance:

        >>> enum_array = household('housing_occupancy_status', period)
        >>> enum_array[0]
        >>> 2  # Encoded value
        >>> enum_array.decode()[0]
        <HousingOccupancyStatus.free_lodger: 'Free lodger'>

        Decoded value: enum item
        """
        return numpy.select(
            [self == item.index for item in self.possible_values],
            list(self.possible_values),
        )

    def decode_to_str(self) -> numpy.str_:
        """
        Return the array of string identifiers corresponding to self.

        For instance:

        >>> enum_array = household('housing_occupancy_status', period)
        >>> enum_array[0]
        >>> 2  # Encoded value
        >>> enum_array.decode_to_str()[0]
        'free_lodger'  # String identifier
        """
        return numpy.select(
            [self == item.index for item in self.possible_values],
            [item.name for item in self.possible_values],
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({str(self.decode())})"

    def __str__(self) -> str:
        return str(self.decode_to_str())



from __future__ import annotations

import copy
import glob
import importlib
import inspect
import logging
import os
import sys
import traceback
import typing
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Type,
    Union,
)

import pkg_resources

from policyengine_core import commons, periods, variables
from policyengine_core.data_structures.reference import Reference
from policyengine_core.entities import Entity
from policyengine_core.errors import (
    VariableNameConflictError,
    VariableNotFoundError,
)
from policyengine_core.parameters import ParameterNode, ParameterNodeAtInstant
from policyengine_core.parameters.operations.homogenize_parameters import (
    homogenize_parameter_structures,
)
from policyengine_core.parameters.operations.interpolate_parameters import (
    interpolate_parameters,
)
from policyengine_core.parameters.operations.propagate_parameter_metadata import (
    propagate_parameter_metadata,
)
from policyengine_core.parameters.operations.uprate_parameters import (
    uprate_parameters,
)
from policyengine_core.periods import Instant, Period
from policyengine_core.populations import GroupPopulation, Population
from policyengine_core.variables import Variable

log = logging.getLogger(__name__)


class TaxBenefitSystem:
    """
    Represents the legislation.

    It stores parameters (values defined for everyone) and variables (values defined for some given entity e.g. a person).

    Attributes:
        parameters: Directory containing the YAML parameter files.

    Args:
        entities: Entities used by the tax benefit system.

    """

    _base_tax_benefit_system: "TaxBenefitSystem" = None
    _parameters_at_instant_cache: Optional[Dict[Any, Any]] = None
    person_key_plural: str = None
    preprocess_parameters: str = None
    baseline: "TaxBenefitSystem" = None  # Baseline tax-benefit system. Used only by reforms. Note: Reforms can be chained.
    cache_blacklist = None
    decomposition_file_path = None

    # The following properties should be specified by country packages.

    entities: List[Entity] = None
    """The entities of the tax and benefit system."""
    variables_dir: str = None
    """Directory containing the Python files defining the variables of the tax and benefit system."""
    parameters_dir: str = None
    """Directory containing the YAML parameter tree."""
    auto_carry_over_input_variables: bool = False
    """Whether to automatically carry over input variables when calculating a variable for a period different from the period of the input variables."""

    def __init__(self, entities: Sequence[Entity] = None) -> None:
        if entities is None:
            entities = self.entities

        if entities is None:
            raise ValueError("TaxBenefitSystems must have entities defined.")

        # TODO: Currently: Don't use a weakref, because they are cleared by Paste (at least) at each call.
        self.parameters: Optional[ParameterNode] = None
        self._parameters_at_instant_cache = {}  # weakref.WeakValueDictionary()
        self.variables: Dict[Any, Any] = {}
        # Tax benefit systems are mutable, so entities (which need to know about our variables) can't be shared among them
        if entities is None or len(entities) == 0:
            raise Exception(
                "A tax and benefit sytem must have at least an entity."
            )
        self.entities = [copy.copy(entity) for entity in entities]
        self.person_entity = [
            entity for entity in self.entities if entity.is_person
        ][0]
        self.group_entities = [
            entity for entity in self.entities if not entity.is_person
        ]
        self.group_entity_keys = [entity.key for entity in self.group_entities]
        for entity in self.entities:
            entity.set_tax_benefit_system(self)

        if self.variables_dir is not None:
            self.add_variables_from_directory(self.variables_dir)

        if self.parameters_dir is not None:
            self.load_parameters(self.parameters_dir)
            self.parameters = homogenize_parameter_structures(
                self.parameters, self.variables
            )
            self.parameters = propagate_parameter_metadata(self.parameters)
            self.parameters = interpolate_parameters(self.parameters)
            self.parameters = uprate_parameters(self.parameters)
            self.parameters = propagate_parameter_metadata(self.parameters)

    @property
    def base_tax_benefit_system(self) -> "TaxBenefitSystem":
        base_tax_benefit_system = self._base_tax_benefit_system
        if base_tax_benefit_system is None:
            baseline = self.baseline
            if baseline is None:
                return self
            self._base_tax_benefit_system = (
                base_tax_benefit_system
            ) = baseline.base_tax_benefit_system
        return base_tax_benefit_system

    def instantiate_entities(self) -> Dict[str, Population]:
        person = self.person_entity
        members = Population(person)
        entities: typing.Dict[Entity.key, Entity] = {person.key: members}

        for entity in self.group_entities:
            entities[entity.key] = GroupPopulation(entity, members)

        return entities

    def load_variable(
        self, variable_class: Type[Variable], update: bool = False
    ) -> Variable:
        name = variable_class.__name__

        # Check if a Variable with the same name is already registered.
        baseline_variable = self.get_variable(name)
        if baseline_variable and not update:
            raise VariableNameConflictError(
                'Variable "{}" is already defined. Use `update_variable` to replace it.'.format(
                    name
                )
            )

        variable = variable_class(baseline_variable=baseline_variable)
        self.variables[variable.name] = variable

        return variable

    def add_variable(self, variable: Variable) -> Variable:
        """Adds an OpenFisca variable to the tax and benefit system.

        Args:
            variable: The variable to add. Must be a subclass of Variable.

        Raises:
            policyengine_core.errors.VariableNameConflictError: if a variable with the same name have previously been added to the tax and benefit system.

        """
        return self.load_variable(variable, update=False)

    def replace_variable(self, variable: str) -> Variable:
        """
        Replaces an existing OpenFisca variable in the tax and benefit system by a new one.

        The new variable must have the same name than the replaced one.

        If no variable with the given name exists in the tax and benefit system, no error will be raised and the new variable will be simply added.

        :param Variable variable: New variable to add. Must be a subclass of Variable.
        """
        name = variable.__name__
        if self.variables.get(name) is not None:
            del self.variables[name]
        self.load_variable(variable, update=False)

    def update_variable(self, variable: str) -> Variable:
        """
        Updates an existing OpenFisca variable in the tax and benefit system.

        All attributes of the updated variable that are not explicitely overridden by the new ``variable`` will stay unchanged.

        The new variable must have the same name than the updated one.

        If no variable with the given name exists in the tax and benefit system, no error will be raised and the new variable will be simply added.

        :param Variable variable: Variable to add. Must be a subclass of Variable.
        """
        return self.load_variable(variable, update=True)

    def add_variables_from_file(self, file_path: str) -> None:
        """
        Adds all OpenFisca variables contained in a given file to the tax and benefit system.
        """
        try:
            file_name = os.path.splitext(os.path.basename(file_path))[0]

            #  As Python remembers loaded modules by name, in order to prevent collisions, we need to make sure that:
            #  - Files with the same name, but located in different directories, have a different module names. Hence the file path hash in the module name.
            #  - The same file, loaded by different tax and benefit systems, has distinct module names. Hence the `id(self)` in the module name.
            module_name = (
                f"{id(self)}_{hash(os.path.abspath(file_path))}_{file_name}"
            )

            try:
                spec = importlib.util.spec_from_file_location(
                    module_name, file_path
                )
                module = importlib.util.module_from_spec(spec)
                sys.modules[module_name] = module
                spec.loader.exec_module(module)
            except NameError as e:
                logging.error(
                    str(e)
                    + ": if this code used to work, this error might be due to a major change in OpenFisca-Core. Checkout the changelog to learn more: <https://github.com/openfisca/openfisca-core/blob/master/CHANGELOG.md>"
                )
                raise
            potential_variables = [
                getattr(module, item)
                for item in dir(module)
                if not item.startswith("__")
            ]
            for pot_variable in potential_variables:
                # We only want to get the module classes defined in this module (not imported)
                if (
                    inspect.isclass(pot_variable)
                    and issubclass(pot_variable, Variable)
                    and pot_variable.__module__ == module_name
                ):
                    self.add_variable(pot_variable)
        except Exception:
            log.error(
                'Unable to load OpenFisca variables from file "{}"'.format(
                    file_path
                )
            )
            raise

    def add_variables_from_directory(self, directory: str) -> None:
        """
        Recursively explores a directory, and adds all OpenFisca variables found there to the tax and benefit system.
        """
        py_files = glob.glob(os.path.join(directory, "*.py"))
        for py_file in py_files:
            self.add_variables_from_file(py_file)
        subdirectories = glob.glob(os.path.join(directory, "*/"))
        for subdirectory in subdirectories:
            self.add_variables_from_directory(subdirectory)

    def add_variables(self, *variables: List[Type[Variable]]):
        """
        Adds a list of OpenFisca Variables to the `TaxBenefitSystem`.

        See also :any:`add_variable`
        """
        for variable in variables:
            self.add_variable(variable)

    def load_extension(self, extension: str) -> None:
        """
        Loads an extension to the tax and benefit system.

        :param str extension: The extension to load. Can be an absolute path pointing to an extension directory, or the name of an OpenFisca extension installed as a pip package.

        """
        # Load extension from installed pip package
        try:
            package = importlib.import_module(extension)
            extension_directory = package.__path__[0]
        except ImportError:
            message = os.linesep.join(
                [
                    traceback.format_exc(),
                    "Error loading extension: `{}` is neither a directory, nor a package.".format(
                        extension
                    ),
                    "Are you sure it is installed in your environment? If so, look at the stack trace above to determine the origin of this error.",
                    "See more at <https://github.com/openfisca/openfisca-extension-template#installing>.",
                ]
            )
            raise ValueError(message)

        self.add_variables_from_directory(extension_directory)
        param_dir = os.path.join(extension_directory, "parameters")
        if os.path.isdir(param_dir):
            extension_parameters = ParameterNode(directory_path=param_dir)
            self.parameters.merge(extension_parameters)

    def apply_reform(self, reform_path: str) -> "TaxBenefitSystem":
        """Generates a new tax and benefit system applying a reform to the tax and benefit system.

        The current tax and benefit system is **not** mutated.

        Args:
            reform_path: The reform to apply. Must respect the format *installed_package.sub_module.reform*

        Returns:
            TaxBenefitSystem: A reformed tax and benefit system.

        Example:

            >>> self.apply_reform('openfisca_france.reforms.inversion_revenus')

        """
        from policyengine_core.reforms import Reform

        try:
            reform_package, reform_name = reform_path.rsplit(".", 1)
        except ValueError:
            raise ValueError(
                "`{}` does not seem to be a path pointing to a reform. A path looks like `some_country_package.reforms.some_reform.`".format(
                    reform_path
                )
            )
        try:
            reform_module = importlib.import_module(reform_package)
        except ImportError:
            message = os.linesep.join(
                [
                    traceback.format_exc(),
                    "Could not import `{}`.".format(reform_package),
                    "Are you sure of this reform module name? If so, look at the stack trace above to determine the origin of this error.",
                ]
            )
            raise ValueError(message)
        reform = getattr(reform_module, reform_name, None)
        if reform is None:
            raise ValueError(
                "{} has no attribute {}".format(reform_package, reform_name)
            )
        if not issubclass(reform, Reform):
            raise ValueError(
                "`{}` does not seem to be a valid Openfisca reform.".format(
                    reform_path
                )
            )

        return reform(self)

    def get_variable(
        self, variable_name: str, check_existence: bool = False
    ) -> Variable:
        """
        Get a variable from the tax and benefit system.

        :param variable_name: Name of the requested variable.
        :param check_existence: If True, raise an error if the requested variable does not exist.
        """
        variables = self.variables
        found = variables.get(variable_name)
        if not found and check_existence:
            raise VariableNotFoundError(variable_name, self)
        return found

    def neutralize_variable(self, variable_name: str) -> None:
        """
        Neutralizes an OpenFisca variable existing in the tax and benefit system.

        A neutralized variable always returns its default value when computed.

        Trying to set inputs for a neutralized variable has no effect except raising a warning.
        """
        self.variables[variable_name] = variables.get_neutralized_variable(
            self.get_variable(variable_name)
        )

    def annualize_variable(
        self, variable_name: str, period: typing.Optional[Period] = None
    ):
        self.variables[variable_name] = variables.get_annualized_variable(
            self.get_variable(variable_name, period)
        )

    def load_parameters(
        self,
        path_to_yaml_dir: str,
    ) -> None:
        """
        Loads the legislation parameter for a directory containing YAML parameters files.

        :param path_to_yaml_dir: Absolute path towards the YAML parameter directory.

        Example:

        >>> self.load_parameters('/path/to/yaml/parameters/dir')
        """

        parameters = ParameterNode(
            "",
            directory_path=path_to_yaml_dir,
        )

        if self.preprocess_parameters is not None:
            parameters = self.preprocess_parameters(parameters)

        self.parameters = parameters

    def _get_baseline_parameters_at_instant(
        self, instant: Instant
    ) -> ParameterNodeAtInstant:
        baseline = self.baseline
        if baseline is None:
            return self.get_parameters_at_instant(instant)
        return baseline._get_baseline_parameters_at_instant(instant)

    def get_parameters_at_instant(
        self, instant: Instant
    ) -> ParameterNodeAtInstant:
        """
        Get the parameters of the legislation at a given instant

        :param instant: :obj:`str` of the format 'YYYY-MM-DD' or :class:`.Instant` instance.
        :returns: The parameters of the legislation at a given instant.
        :rtype: :class:`.ParameterNodeAtInstant`
        """
        if isinstance(instant, Period):
            instant = instant.start
        elif isinstance(instant, (str, int)):
            instant = periods.instant(instant)
        else:
            assert isinstance(
                instant, Instant
            ), "Expected an Instant (e.g. Instant((2017, 1, 1)) ). Got: {}.".format(
                instant
            )

        parameters_at_instant = self._parameters_at_instant_cache.get(instant)
        if parameters_at_instant is None and self.parameters is not None:
            parameters_at_instant = self.parameters.get_at_instant(
                str(instant)
            )
            self._parameters_at_instant_cache[instant] = parameters_at_instant
        return parameters_at_instant

    def get_package_metadata(self) -> dict:
        """
        Gets metatada relative to the country package the tax and benefit system is built from.

        :returns: Country package metadata
        :rtype: dict

        Example:

        >>> tax_benefit_system.get_package_metadata()
        >>> {
        >>>    'location': '/path/to/dir/containing/package',
        >>>    'name': 'openfisca-france',
        >>>    'repository_url': 'https://github.com/openfisca/openfisca-france',
        >>>    'version': '17.2.0'
        >>>    }
        """
        # Handle reforms
        if self.baseline:
            return self.baseline.get_package_metadata()

        fallback_metadata = {
            "name": self.__class__.__name__,
            "version": "",
            "repository_url": "",
            "location": "",
        }

        module = inspect.getmodule(self)
        if not module.__package__:
            return fallback_metadata
        package_name = module.__package__.split(".")[0]
        try:
            distribution = pkg_resources.get_distribution(package_name)
        except pkg_resources.DistributionNotFound:
            return fallback_metadata

        location = (
            inspect.getsourcefile(module).split(package_name)[0].rstrip("/")
        )

        home_page_metadatas = [
            metadata.split(":", 1)[1].strip(" ")
            for metadata in distribution._get_metadata(distribution.PKG_INFO)
            if "Home-page" in metadata
        ]
        repository_url = home_page_metadatas[0] if home_page_metadatas else ""
        return {
            "name": distribution.key,
            "version": distribution.version,
            "repository_url": repository_url,
            "location": location,
        }

    def get_variables(self, entity: Entity = None) -> dict:
        """
        Gets all variables contained in a tax and benefit system.

        :param Entity entity: If set, returns only the variable defined for the given entity.

        :returns: A dictionary, indexed by variable names.
        :rtype: dict

        """
        if not entity:
            return self.variables
        else:
            return {
                variable_name: variable
                for variable_name, variable in self.variables.items()
                # TODO - because entities are copied (see constructor) they can't be compared
                if variable.entity.key == entity.key
            }

    def clone(self) -> "TaxBenefitSystem":
        new = commons.empty_clone(self)
        new_dict = new.__dict__

        for key, value in self.__dict__.items():
            if key not in (
                "parameters",
                "_parameters_at_instant_cache",
                "variables",
            ):
                new_dict[key] = value
        for entity in new_dict["entities"]:
            entity.set_tax_benefit_system(new)

        new_dict["parameters"] = self.parameters.clone()
        new_dict["_parameters_at_instant_cache"] = {}
        new_dict["variables"] = self.variables.copy()
        return new

    def entities_plural(self) -> dict:
        return {entity.plural for entity in self.entities}

    def entities_by_singular(self) -> dict:
        return {entity.key: entity for entity in self.entities}

    def test(self, paths: str, verbose: bool = False) -> None:
        from policyengine_core.tools.test_runner import (
            run_tests,
        )  # Import here to avoid circular dependency.

        run_tests(self, paths, options=dict(verbose=verbose))



import calendar
import datetime

from policyengine_core import periods
from policyengine_core.periods import config


class Instant(tuple):
    def __repr__(self) -> str:
        """
        Transform instant to to its Python representation as a string.

        >>> repr(instant(2014))
        'Instant((2014, 1, 1))'
        >>> repr(instant('2014-2'))
        'Instant((2014, 2, 1))'
        >>> repr(instant('2014-2-3'))
        'Instant((2014, 2, 3))'
        """
        return "{}({})".format(
            self.__class__.__name__, super(Instant, self).__repr__()
        )

    def __str__(self) -> str:
        """
        Transform instant to a string.

        >>> str(instant(2014))
        '2014-01-01'
        >>> str(instant('2014-2'))
        '2014-02-01'
        >>> str(instant('2014-2-3'))
        '2014-02-03'

        """
        instant_str = config.str_by_instant_cache.get(self)
        if instant_str is None:
            config.str_by_instant_cache[
                self
            ] = instant_str = self.date.isoformat()
        return instant_str

    @property
    def date(self) -> datetime.date:
        """
        Convert instant to a date.

        >>> instant(2014).date
        datetime.date(2014, 1, 1)
        >>> instant('2014-2').date
        datetime.date(2014, 2, 1)
        >>> instant('2014-2-3').date
        datetime.date(2014, 2, 3)
        """
        instant_date = config.date_by_instant_cache.get(self)
        if instant_date is None:
            config.date_by_instant_cache[self] = instant_date = datetime.date(
                *self
            )
        return instant_date

    @property
    def day(self) -> int:
        """
        Extract day from instant.

        >>> instant(2014).day
        1
        >>> instant('2014-2').day
        1
        >>> instant('2014-2-3').day
        3
        """
        return self[2]

    @property
    def month(self) -> int:
        """
        Extract month from instant.

        >>> instant(2014).month
        1
        >>> instant('2014-2').month
        2
        >>> instant('2014-2-3').month
        2
        """
        return self[1]

    def period(self, unit: str, size: int = 1):
        """
        Create a new period starting at instant.

        >>> instant(2014).period('month')
        Period(('month', Instant((2014, 1, 1)), 1))
        >>> instant('2014-2').period('year', 2)
        Period(('year', Instant((2014, 2, 1)), 2))
        >>> instant('2014-2-3').period('day', size = 2)
        Period(('day', Instant((2014, 2, 3)), 2))
        """
        assert unit in (
            config.DAY,
            config.MONTH,
            config.YEAR,
        ), "Invalid unit: {} of type {}".format(unit, type(unit))
        assert (
            isinstance(size, int) and size >= 1
        ), "Invalid size: {} of type {}".format(size, type(size))
        return periods.Period((unit, self, size))

    def offset(self, offset: int, unit: str) -> "Instant":
        """
        Increment (or decrement) the given instant with offset units.

        >>> instant(2014).offset(1, 'day')
        Instant((2014, 1, 2))
        >>> instant(2014).offset(1, 'month')
        Instant((2014, 2, 1))
        >>> instant(2014).offset(1, 'year')
        Instant((2015, 1, 1))

        >>> instant('2014-1-31').offset(1, 'day')
        Instant((2014, 2, 1))
        >>> instant('2014-1-31').offset(1, 'month')
        Instant((2014, 2, 28))
        >>> instant('2014-1-31').offset(1, 'year')
        Instant((2015, 1, 31))

        >>> instant('2011-2-28').offset(1, 'day')
        Instant((2011, 3, 1))
        >>> instant('2011-2-28').offset(1, 'month')
        Instant((2011, 3, 28))
        >>> instant('2012-2-29').offset(1, 'year')
        Instant((2013, 2, 28))

        >>> instant(2014).offset(-1, 'day')
        Instant((2013, 12, 31))
        >>> instant(2014).offset(-1, 'month')
        Instant((2013, 12, 1))
        >>> instant(2014).offset(-1, 'year')
        Instant((2013, 1, 1))

        >>> instant('2011-3-1').offset(-1, 'day')
        Instant((2011, 2, 28))
        >>> instant('2011-3-31').offset(-1, 'month')
        Instant((2011, 2, 28))
        >>> instant('2012-2-29').offset(-1, 'year')
        Instant((2011, 2, 28))

        >>> instant('2014-1-30').offset(3, 'day')
        Instant((2014, 2, 2))
        >>> instant('2014-10-2').offset(3, 'month')
        Instant((2015, 1, 2))
        >>> instant('2014-1-1').offset(3, 'year')
        Instant((2017, 1, 1))

        >>> instant(2014).offset(-3, 'day')
        Instant((2013, 12, 29))
        >>> instant(2014).offset(-3, 'month')
        Instant((2013, 10, 1))
        >>> instant(2014).offset(-3, 'year')
        Instant((2011, 1, 1))

        >>> instant(2014).offset('first-of', 'month')
        Instant((2014, 1, 1))
        >>> instant('2014-2').offset('first-of', 'month')
        Instant((2014, 2, 1))
        >>> instant('2014-2-3').offset('first-of', 'month')
        Instant((2014, 2, 1))

        >>> instant(2014).offset('first-of', 'year')
        Instant((2014, 1, 1))
        >>> instant('2014-2').offset('first-of', 'year')
        Instant((2014, 1, 1))
        >>> instant('2014-2-3').offset('first-of', 'year')
        Instant((2014, 1, 1))

        >>> instant(2014).offset('last-of', 'month')
        Instant((2014, 1, 31))
        >>> instant('2014-2').offset('last-of', 'month')
        Instant((2014, 2, 28))
        >>> instant('2012-2-3').offset('last-of', 'month')
        Instant((2012, 2, 29))

        >>> instant(2014).offset('last-of', 'year')
        Instant((2014, 12, 31))
        >>> instant('2014-2').offset('last-of', 'year')
        Instant((2014, 12, 31))
        >>> instant('2014-2-3').offset('last-of', 'year')
        Instant((2014, 12, 31))
        """
        year, month, day = self
        assert unit in (
            config.DAY,
            config.MONTH,
            config.YEAR,
        ), "Invalid unit: {} of type {}".format(unit, type(unit))
        if offset == "first-of":
            if unit == config.MONTH:
                day = 1
            elif unit == config.YEAR:
                month = 1
                day = 1
        elif offset == "last-of":
            if unit == config.MONTH:
                day = calendar.monthrange(year, month)[1]
            elif unit == config.YEAR:
                month = 12
                day = 31
        else:
            assert isinstance(
                offset, int
            ), "Invalid offset: {} of type {}".format(offset, type(offset))
            if unit == config.DAY:
                day += offset
                if offset < 0:
                    while day < 1:
                        month -= 1
                        if month == 0:
                            year -= 1
                            month = 12
                        day += calendar.monthrange(year, month)[1]
                elif offset > 0:
                    month_last_day = calendar.monthrange(year, month)[1]
                    while day > month_last_day:
                        month += 1
                        if month == 13:
                            year += 1
                            month = 1
                        day -= month_last_day
                        month_last_day = calendar.monthrange(year, month)[1]
            elif unit == config.MONTH:
                month += offset
                if offset < 0:
                    while month < 1:
                        year -= 1
                        month += 12
                elif offset > 0:
                    while month > 12:
                        year += 1
                        month -= 12
                month_last_day = calendar.monthrange(year, month)[1]
                if day > month_last_day:
                    day = month_last_day
            elif unit == config.YEAR:
                year += offset
                # Handle february month of leap year.
                month_last_day = calendar.monthrange(year, month)[1]
                if day > month_last_day:
                    day = month_last_day

        return self.__class__((year, month, day))

    @property
    def year(self) -> int:
        """
        Extract year from instant.

        >>> instant(2014).year
        2014
        >>> instant('2014-2').year
        2014
        >>> instant('2014-2-3').year
        2014
        """
        return self[0]



import os
import warnings
from typing import TYPE_CHECKING, Any, List

import numpy
import psutil
from numpy.typing import ArrayLike

from policyengine_core import commons, periods, tools
from policyengine_core.data_storage import InMemoryStorage, OnDiskStorage
from policyengine_core.enums import Enum
from policyengine_core.errors import PeriodMismatchError
from policyengine_core.periods import Period

if TYPE_CHECKING:
    from policyengine_core.populations import Population
    from policyengine_core.variables import Variable


class Holder:
    """
    A holder keeps tracks of a variable values after they have been calculated, or set as an input.
    """

    def __init__(self, variable: "Variable", population: "Population"):
        self.population = population
        self.variable = variable
        self.simulation = population.simulation
        self._memory_storage = InMemoryStorage(
            is_eternal=(self.variable.definition_period == periods.ETERNITY)
        )

        # By default, do not activate on-disk storage, or variable dropping
        self._disk_storage = None
        self._on_disk_storable = False
        self._do_not_store = False
        if self.simulation and self.simulation.memory_config:
            if (
                self.variable.name
                not in self.simulation.memory_config.priority_variables
            ):
                self._disk_storage = self.create_disk_storage()
                self._on_disk_storable = True
            if (
                self.variable.name
                in self.simulation.memory_config.variables_to_drop
            ):
                self._do_not_store = True

    def clone(self, population: "Population") -> "Holder":
        """
        Copy the holder just enough to be able to run a new simulation without modifying the original simulation.
        """
        new = commons.empty_clone(self)
        new_dict = new.__dict__

        for key, value in self.__dict__.items():
            if key not in ("population", "formula", "simulation"):
                new_dict[key] = value

        new_dict["population"] = population
        new_dict["simulation"] = population.simulation

        return new

    def create_disk_storage(
        self, directory: str = None, preserve: bool = False
    ) -> OnDiskStorage:
        if directory is None:
            directory = self.simulation.data_storage_dir
        storage_dir = os.path.join(directory, self.variable.name)
        if not os.path.isdir(storage_dir):
            os.mkdir(storage_dir)
        return OnDiskStorage(
            storage_dir,
            is_eternal=(self.variable.definition_period == periods.ETERNITY),
            preserve_storage_dir=preserve,
        )

    def delete_arrays(self, period: Period = None) -> None:
        """
        If ``period`` is ``None``, remove all known values of the variable.

        If ``period`` is not ``None``, only remove all values for any period included in period (e.g. if period is "2017", values for "2017-01", "2017-07", etc. would be removed)
        """

        self._memory_storage.delete(period)
        if self._disk_storage:
            self._disk_storage.delete(period)

    def get_array(self, period: Period) -> ArrayLike:
        """
        Get the value of the variable for the given period.

        If the value is not known, return ``None``.
        """
        if self.variable.is_neutralized:
            return self.default_array()
        value = self._memory_storage.get(period)
        if value is not None:
            return value
        if self._disk_storage:
            return self._disk_storage.get(period)

    def get_memory_usage(self) -> dict:
        """
        Get data about the virtual memory usage of the holder.

        :returns: Memory usage data
        :rtype: dict

        Example:

        >>> holder.get_memory_usage()
        >>> {
        >>>    'nb_arrays': 12,  # The holder contains the variable values for 12 different periods
        >>>    'nb_cells_by_array': 100, # There are 100 entities (e.g. persons) in our simulation
        >>>    'cell_size': 8,  # Each value takes 8B of memory
        >>>    'dtype': dtype('float64')  # Each value is a float 64
        >>>    'total_nb_bytes': 10400  # The holder uses 10.4kB of virtual memory
        >>>    'nb_requests': 24  # The variable has been computed 24 times
        >>>    'nb_requests_by_array': 2  # Each array stored has been on average requested twice
        >>>    }
        """

        usage = dict(
            nb_cells_by_array=self.population.count,
            dtype=self.variable.dtype,
        )

        usage.update(self._memory_storage.get_memory_usage())

        if self.simulation.trace:
            nb_requests = self.simulation.tracer.get_nb_requests(
                self.variable.name
            )
            usage.update(
                dict(
                    nb_requests=nb_requests,
                    nb_requests_by_array=nb_requests
                    / float(usage["nb_arrays"])
                    if usage["nb_arrays"] > 0
                    else numpy.nan,
                )
            )

        return usage

    def get_known_periods(self) -> List[Period]:
        """
        Get the list of periods the variable value is known for.
        """

        return list(self._memory_storage.get_known_periods()) + list(
            (
                self._disk_storage.get_known_periods()
                if self._disk_storage
                else []
            )
        )

    def set_input(self, period: Period, array: ArrayLike) -> None:
        """
        Set a variable's value (``array``) for a given period (``period``)

        :param array: the input value for the variable
        :param period: the period at which the value is setted

        Example :

        >>> holder.set_input([12, 14], '2018-04')
        >>> holder.get_array('2018-04')
        >>> [12, 14]


        If a ``set_input`` property has been set for the variable, this method may accept inputs for periods not matching the ``definition_period`` of the variable. To read more about this, check the `documentation <https://openfisca.org/doc/coding-the-legislation/35_periods.html#set-input-automatically-process-variable-inputs-defined-for-periods-not-matching-the-definition-period>`_.
        """

        period = periods.period(period)
        if (
            period.unit == periods.ETERNITY
            and self.variable.definition_period != periods.ETERNITY
        ):
            error_message = os.linesep.join(
                [
                    "Unable to set a value for variable {0} for periods.ETERNITY.",
                    "{0} is only defined for {1}s. Please adapt your input.",
                ]
            ).format(self.variable.name, self.variable.definition_period)
            raise PeriodMismatchError(
                self.variable.name,
                period,
                self.variable.definition_period,
                error_message,
            )
        if self.variable.is_neutralized:
            warning_message = "You cannot set a value for the variable {}, as it has been neutralized. The value you provided ({}) will be ignored.".format(
                self.variable.name, array
            )
            return warnings.warn(warning_message, Warning)
        if self.variable.value_type in (float, int) and isinstance(array, str):
            array = tools.eval_expression(array)
        if self.variable.set_input:
            return self.variable.set_input(self, period, array)
        return self._set(period, array)

    def _to_array(self, value: Any) -> ArrayLike:
        if not isinstance(value, numpy.ndarray):
            value = numpy.asarray(value)
        if value.ndim == 0:
            # 0-dim arrays are casted to scalar when they interact with float. We don't want that.
            value = value.reshape(1)
        if len(value) != self.population.count:
            raise ValueError(
                'Unable to set value "{}" for variable "{}", as its length is {} while there are {} {} in the simulation.'.format(
                    value,
                    self.variable.name,
                    len(value),
                    self.population.count,
                    self.population.entity.plural,
                )
            )
        if self.variable.value_type == Enum:
            value = self.variable.possible_values.encode(value)
        if value.dtype != self.variable.dtype:
            try:
                value = value.astype(self.variable.dtype)
            except ValueError:
                raise ValueError(
                    'Unable to set value "{}" for variable "{}", as the variable dtype "{}" does not match the value dtype "{}".'.format(
                        value,
                        self.variable.name,
                        self.variable.dtype,
                        value.dtype,
                    )
                )
        return value

    def _set(self, period: Period, value: ArrayLike) -> None:
        value = self._to_array(value)
        if self.variable.definition_period != periods.ETERNITY:
            if period is None:
                raise ValueError(
                    "A period must be specified to set values, except for variables with periods.ETERNITY as as period_definition."
                )
            if (
                self.variable.definition_period != period.unit
                or period.size > 1
            ):
                name = self.variable.name
                period_size_adj = (
                    f"{period.unit}"
                    if (period.size == 1)
                    else f"{period.size}-{period.unit}s"
                )
                error_message = os.linesep.join(
                    [
                        f'Unable to set a value for variable "{name}" for {period_size_adj}-long period "{period}".',
                        f'"{name}" can only be set for one {self.variable.definition_period} at a time. Please adapt your input.',
                        f'If you are the maintainer of "{name}", you can consider adding it a set_input attribute to enable automatic period casting.',
                    ]
                )

                raise PeriodMismatchError(
                    self.variable.name,
                    period,
                    self.variable.definition_period,
                    error_message,
                )

        should_store_on_disk = (
            self._on_disk_storable
            and self._memory_storage.get(period) is None
            and psutil.virtual_memory().percent  # If there is already a value in memory, replace it and don't put a new value in the disk storage
            >= self.simulation.memory_config.max_memory_occupation_pc
        )

        if should_store_on_disk:
            self._disk_storage.put(value, period)
        else:
            self._memory_storage.put(value, period)

    def put_in_cache(self, value: ArrayLike, period: Period) -> None:
        if self._do_not_store:
            return

        if (
            self.simulation.opt_out_cache
            and self.simulation.tax_benefit_system.cache_blacklist
            and self.variable.name
            in self.simulation.tax_benefit_system.cache_blacklist
        ):
            return

        self._set(period, value)

    def default_array(self) -> ArrayLike:
        """
        Return a new array of the appropriate length for the entity, filled with the variable default values.
        """

        return self.variable.default_array(self.population.count)



from .entity_to_person_projector import EntityToPersonProjector
from .first_person_to_entity_projector import FirstPersonToEntityProjector
from .helpers import get_projector_from_shortcut, projectable
from .projector import Projector
from .unique_role_to_entity_projector import UniqueRoleToEntityProjector



import argparse
import sys
import warnings

from policyengine_core.scripts import add_tax_benefit_system_arguments

"""
    Define the `openfisca` command line interface.
"""


def get_parser():
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(
        help="Available commands", dest="command"
    )
    subparsers.required = (
        True  # Can be added as an argument of add_subparsers in Python 3
    )

    def build_test_parser(parser):
        parser.add_argument(
            "path",
            help="paths (files or directories) of tests to execute",
            nargs="+",
        )
        parser = add_tax_benefit_system_arguments(parser)
        parser.add_argument(
            "-n",
            "--name_filter",
            default=None,
            help="partial name of tests to execute. Only tests with the given name_filter in their name, file name, or keywords will be run.",
        )
        parser.add_argument(
            "-p",
            "--pdb",
            action="store_true",
            default=False,
            help="drop into debugger on failures or errors",
        )
        parser.add_argument(
            "--performance-graph",
            "--performance",
            action="store_true",
            default=False,
            help="output a performance graph in a 'performance_graph.html' file",
        )
        parser.add_argument(
            "--performance-tables",
            action="store_true",
            default=False,
            help="output performance CSV tables",
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            default=False,
            help="increase output verbosity. If specified, output the entire calculation trace.",
        )
        parser.add_argument(
            "-a",
            "--aggregate",
            action="store_true",
            default=False,
            help="increase output verbosity to aggregate. If specified, output the avg, max, and min values of the calculation trace. This flag has no effect without --verbose.",
        )
        parser.add_argument(
            "-d",
            "--max-depth",
            type=int,
            default=None,
            help="set maximal verbosity depth. If specified, output the calculation trace up to the provided depth. This flag has no effect without --verbose.",
        )
        parser.add_argument(
            "-o",
            "--only-variables",
            nargs="*",
            default=None,
            help="variables to test. If specified, only test the given variables.",
        )
        parser.add_argument(
            "-i",
            "--ignore-variables",
            nargs="*",
            default=None,
            help="variables to ignore. If specified, do not test the given variables.",
        )

        return parser

    def build_data_parser(parser):
        parser.add_argument("dataset", help="The dataset to focus on.")
        parser = add_tax_benefit_system_arguments(parser, country_only=True)
        parser.add_argument(
            "action",
            choices=["generate", "download", "upload", "remove", "list"],
            help="The action to perform. Pass any additional options as keyword arguments.",
        )

        return parser

    parser_test = subparsers.add_parser(
        "test", help="Run OpenFisca YAML tests"
    )
    parser_test = build_test_parser(parser_test)

    parser_data = subparsers.add_parser("data", help="Manage OpenFisca data")
    parser_data = build_data_parser(parser_data)

    return parser


def main():

    parser = get_parser()

    args, _ = parser.parse_known_args()

    if args.command == "test":
        from policyengine_core.scripts.run_test import main

        return sys.exit(main(parser))

    if args.command == "data":
        from policyengine_core.scripts.run_data import main

        return sys.exit(main(parser))


if __name__ == "__main__":
    sys.exit(main())



from policyengine_core.country_template.entities import *
from policyengine_core.model_api import *


class person_id(Variable):
    value_type = int
    entity = Person
    definition_period = ETERNITY
    label = "Person ID"


class household_id(Variable):
    value_type = int
    entity = Household
    definition_period = ETERNITY
    label = "Household ID"


class household_weight(Variable):
    value_type = float
    entity = Household
    definition_period = YEAR
    label = "Household weight"


class person_weight(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "Person weight"

    def formula(person, period):
        return person.household("household_weight", period)


class person_household_role(Variable):
    value_type = str
    entity = Person
    definition_period = ETERNITY
    label = "Person household role"


class person_household_id(Variable):
    value_type = int
    entity = Person
    definition_period = ETERNITY
    label = "Person household ID"



from typing import TYPE_CHECKING, Any, Callable

import numpy
from numpy.typing import ArrayLike

from policyengine_core import projectors
from policyengine_core.entities import Entity, Role
from policyengine_core.enums import EnumArray
from policyengine_core.populations.population import Population

if TYPE_CHECKING:
    from policyengine_core.simulations import Simulation


class GroupPopulation(Population):
    def __init__(self, entity: Entity, members: Population):
        super().__init__(entity)
        self.members: Population = members
        self._members_entity_id: ArrayLike = None
        self._members_role: ArrayLike = None
        self._members_position: ArrayLike = None
        self._ordered_members_map = None

    def clone(
        self, simulation: "Simulation", members: Population
    ) -> "GroupPopulation":
        result = GroupPopulation(self.entity, members)
        result.simulation = simulation
        result._holders = {
            variable: holder.clone(self)
            for (variable, holder) in self._holders.items()
        }
        result.count = self.count
        result.ids = self.ids
        result._members_entity_id = self._members_entity_id
        result._members_role = self._members_role
        result._members_position = self._members_position
        result._ordered_members_map = self._ordered_members_map
        return result

    @property
    def members_position(self) -> ArrayLike:
        if (
            self._members_position is None
            and self.members_entity_id is not None
        ):
            # We could use self.count and self.members.count , but with the current initilization, we are not sure count will be set before members_position is called
            nb_entities = numpy.max(self.members_entity_id) + 1
            nb_persons = len(self.members_entity_id)
            self._members_position = numpy.empty_like(self.members_entity_id)
            counter_by_entity = numpy.zeros(nb_entities)
            for k in range(nb_persons):
                entity_index = self.members_entity_id[k]
                self._members_position[k] = counter_by_entity[entity_index]
                counter_by_entity[entity_index] += 1

        return self._members_position

    @members_position.setter
    def members_position(self, members_position: ArrayLike) -> None:
        self._members_position = members_position

    @property
    def members_entity_id(self) -> ArrayLike:
        return self._members_entity_id

    @members_entity_id.setter
    def members_entity_id(self, members_entity_id: ArrayLike) -> None:
        self._members_entity_id = members_entity_id

    @property
    def members_role(self) -> ArrayLike:
        if self._members_role is None:
            default_role = self.entity.flattened_roles[0]
            self._members_role = numpy.repeat(
                default_role, len(self.members_entity_id)
            )
        return self._members_role

    @members_role.setter
    def members_role(self, members_role: ArrayLike):
        if members_role is not None:
            self._members_role = numpy.array(members_role)

    @property
    def ordered_members_map(self) -> ArrayLike:
        """
        Mask to group the persons by entity
        This function only caches the map value, to see what the map is used for, see value_nth_person method.
        """
        if self._ordered_members_map is None:
            self._ordered_members_map = numpy.argsort(self.members_entity_id)
        return self._ordered_members_map

    def get_role(self, role_name: str) -> Role:
        return next(
            (
                role
                for role in self.entity.flattened_roles
                if role.key == role_name
            ),
            None,
        )

    #  Aggregation persons -> entity

    @projectors.projectable
    def sum(self, array: ArrayLike, role: Role = None) -> ArrayLike:
        """
        Return the sum of ``array`` for the members of the entity.

        ``array`` must have the dimension of the number of persons in the simulation

        If ``role`` is provided, only the entity member with the given role are taken into account.

        Example:

        >>> salaries = household.members('salary', '2018-01')  # e.g. [2000, 1500, 0, 0, 0]
        >>> household.sum(salaries)
        >>> array([3500])
        """
        self.entity.check_role_validity(role)
        self.members.check_array_compatible_with_entity(array)
        if role is not None:
            role_filter = self.members.has_role(role)
            return numpy.bincount(
                self.members_entity_id[role_filter],
                weights=array[role_filter],
                minlength=self.count,
            )
        else:
            return numpy.bincount(self.members_entity_id, weights=array)

    @projectors.projectable
    def any(self, array: ArrayLike, role: Role = None) -> ArrayLike:
        """
        Return ``True`` if ``array`` is ``True`` for any members of the entity.

        ``array`` must have the dimension of the number of persons in the simulation

        If ``role`` is provided, only the entity member with the given role are taken into account.

        Example:

        >>> salaries = household.members('salary', '2018-01')  # e.g. [2000, 1500, 0, 0, 0]
        >>> household.any(salaries >= 1800)
        >>> array([True])
        """
        sum_in_entity = self.sum(array, role=role)
        return sum_in_entity > 0

    @projectors.projectable
    def reduce(
        self,
        array: ArrayLike,
        reducer: Callable,
        neutral_element: Any,
        role: Role = None,
    ) -> ArrayLike:
        self.members.check_array_compatible_with_entity(array)
        self.entity.check_role_validity(role)
        position_in_entity = self.members_position
        role_filter = self.members.has_role(role) if role is not None else True
        filtered_array = numpy.where(role_filter, array, neutral_element)

        result = self.filled_array(
            neutral_element
        )  # Neutral value that will be returned if no one with the given role exists.

        # We loop over the positions in the entity
        # Looping over the entities is tempting, but potentielly slow if there are a lot of entities
        biggest_entity_size = numpy.max(position_in_entity) + 1

        for p in range(biggest_entity_size):
            values = self.value_nth_person(
                p, filtered_array, default=neutral_element
            )
            result = reducer(result, values)

        return result

    @projectors.projectable
    def all(self, array: ArrayLike, role: Role = None) -> ArrayLike:
        """
        Return ``True`` if ``array`` is ``True`` for all members of the entity.

        ``array`` must have the dimension of the number of persons in the simulation

        If ``role`` is provided, only the entity member with the given role are taken into account.

        Example:

        >>> salaries = household.members('salary', '2018-01')  # e.g. [2000, 1500, 0, 0, 0]
        >>> household.all(salaries >= 1800)
        >>> array([False])
        """
        return self.reduce(
            array, reducer=numpy.logical_and, neutral_element=True, role=role
        )

    @projectors.projectable
    def max(self, array: ArrayLike, role: Role = None) -> ArrayLike:
        """
        Return the maximum value of ``array`` for the entity members.

        ``array`` must have the dimension of the number of persons in the simulation

        If ``role`` is provided, only the entity member with the given role are taken into account.

        Example:

        >>> salaries = household.members('salary', '2018-01')  # e.g. [2000, 1500, 0, 0, 0]
        >>> household.max(salaries)
        >>> array([2000])
        """
        return self.reduce(
            array,
            reducer=numpy.maximum,
            neutral_element=-numpy.infty,
            role=role,
        )

    @projectors.projectable
    def min(self, array: ArrayLike, role: Role = None) -> ArrayLike:
        """
        Return the minimum value of ``array`` for the entity members.

        ``array`` must have the dimension of the number of persons in the simulation

        If ``role`` is provided, only the entity member with the given role are taken into account.

        Example:

        >>> salaries = household.members('salary', '2018-01')  # e.g. [2000, 1500, 0, 0, 0]
        >>> household.min(salaries)
        >>> array([0])
        >>> household.min(salaries, role = Household.PARENT)  # Assuming the 1st two persons are parents
        >>> array([1500])
        """
        return self.reduce(
            array,
            reducer=numpy.minimum,
            neutral_element=numpy.infty,
            role=role,
        )

    @projectors.projectable
    def nb_persons(self, role: Role = None) -> ArrayLike:
        """
        Returns the number of persons contained in the entity.

        If ``role`` is provided, only the entity member with the given role are taken into account.
        """
        if role:
            if role.subroles:
                role_condition = numpy.logical_or.reduce(
                    [self.members_role == subrole for subrole in role.subroles]
                )
            else:
                role_condition = self.members_role == role
            return self.sum(role_condition)
        else:
            return numpy.bincount(self.members_entity_id)

    # Projection person -> entity

    @projectors.projectable
    def value_from_person(
        self, array: ArrayLike, role: Role, default: Any = 0
    ) -> ArrayLike:
        """
        Get the value of ``array`` for the person with the unique role ``role``.

        ``array`` must have the dimension of the number of persons in the simulation

        If such a person does not exist, return ``default`` instead

        The result is a vector which dimension is the number of entities
        """
        self.entity.check_role_validity(role)
        if role.max != 1:
            raise Exception(
                "You can only use value_from_person with a role that is unique in {}. Role {} is not unique.".format(
                    self.key, role.key
                )
            )
        self.members.check_array_compatible_with_entity(array)
        members_map = self.ordered_members_map
        result = self.filled_array(default, dtype=array.dtype)
        if isinstance(array, EnumArray):
            result = EnumArray(result, array.possible_values)
        role_filter = self.members.has_role(role)
        entity_filter = self.any(role_filter)

        result[entity_filter] = array[members_map][role_filter[members_map]]

        return result

    @projectors.projectable
    def value_nth_person(
        self, n: int, array: ArrayLike, default: Any = 0
    ) -> ArrayLike:
        """
        Get the value of array for the person whose position in the entity is n.

        Note that this position is arbitrary, and that members are not sorted.

        If the nth person does not exist, return  ``default`` instead.

        The result is a vector which dimension is the number of entities.
        """
        self.members.check_array_compatible_with_entity(array)
        positions = self.members_position
        nb_persons_per_entity = self.nb_persons()
        members_map = self.ordered_members_map
        result = self.filled_array(default, dtype=array.dtype)
        # For households that have at least n persons, set the result as the value of criteria for the person for which the position is n.
        # The map is needed b/c the order of the nth persons of each household in the persons vector is not necessarily the same than the household order.
        result[nb_persons_per_entity > n] = array[members_map][
            positions[members_map] == n
        ]

        if isinstance(array, EnumArray):
            result = EnumArray(result, array.possible_values)

        return result

    @projectors.projectable
    def value_from_first_person(self, array: ArrayLike):
        return self.value_nth_person(0, array)

    # Projection entity -> person(s)

    def project(self, array: ArrayLike, role: Role = None) -> ArrayLike:
        self.check_array_compatible_with_entity(array)
        self.entity.check_role_validity(role)
        if role is None:
            return array[self.members_entity_id]
        else:
            role_condition = self.members.has_role(role)
            return numpy.where(
                role_condition, array[self.members_entity_id], 0
            )



from typing import Dict, Union

import numpy
from numpy.typing import ArrayLike

from policyengine_core import periods
from policyengine_core.periods import Period


class InMemoryStorage:
    """
    Low-level class responsible for storing and retrieving calculated vectors in memory
    """

    _arrays: Dict[Period, ArrayLike]
    is_eternal: bool

    def __init__(self, is_eternal: bool):
        self._arrays = {}
        self.is_eternal = is_eternal

    def get(self, period: Period) -> ArrayLike:
        if self.is_eternal:
            period = periods.period(periods.ETERNITY)
        period = periods.period(period)

        values = self._arrays.get(period)
        if values is None:
            return None
        return values

    def put(self, value: ArrayLike, period: Period) -> None:
        if self.is_eternal:
            period = periods.period(periods.ETERNITY)
        period = periods.period(period)

        self._arrays[period] = value

    def delete(self, period: Period = None) -> None:
        if period is None:
            self._arrays = {}
            return

        if self.is_eternal:
            period = periods.period(periods.ETERNITY)
        period = periods.period(period)

        self._arrays = {
            period_item: value
            for period_item, value in self._arrays.items()
            if not period.contains(period_item)
        }

    def get_known_periods(self) -> list:
        return list(self._arrays.keys())

    def get_memory_usage(self) -> dict:
        if not self._arrays:
            return dict(
                nb_arrays=0,
                total_nb_bytes=0,
                cell_size=numpy.nan,
            )

        nb_arrays = len(self._arrays)
        array = next(iter(self._arrays.values()))
        return dict(
            nb_arrays=nb_arrays,
            total_nb_bytes=array.nbytes * nb_arrays,
            cell_size=array.itemsize,
        )



from __future__ import annotations

import typing
from typing import List, Optional, Union

import numpy

from policyengine_core.enums import EnumArray

from .. import tracers

if typing.TYPE_CHECKING:
    from numpy.typing import ArrayLike

    Array = Union[EnumArray, ArrayLike]


class ComputationLog:

    _full_tracer: tracers.FullTracer

    def __init__(self, full_tracer: tracers.FullTracer) -> None:
        self._full_tracer = full_tracer

    def display(
        self,
        value: Optional[Array],
    ) -> str:
        if isinstance(value, EnumArray):
            value = value.decode_to_str()

        return numpy.array2string(value, max_line_width=float("inf"))

    def lines(
        self,
        aggregate: bool = False,
        max_depth: Optional[int] = None,
    ) -> List[str]:
        depth = 1

        lines_by_tree = [
            self._get_node_log(node, depth, aggregate, max_depth)
            for node in self._full_tracer.trees
        ]

        return self._flatten(lines_by_tree)

    def print_log(self, aggregate=False, max_depth=None) -> None:
        """
        Print the computation log of a simulation.

        If ``aggregate`` is ``False`` (default), print the value of each
        computed vector.

        If ``aggregate`` is ``True``, only print the minimum, maximum, and
        average value of each computed vector.

        This mode is more suited for simulations on a large population.

        If ``max_depth`` is ``None`` (default), print the entire computation.

        If ``max_depth`` is set, for example to ``3``, only print computed
        vectors up to a depth of ``max_depth``.
        """
        for line in self.lines(aggregate, max_depth):
            print(line)

    def _get_node_log(
        self,
        node: tracers.TraceNode,
        depth: int,
        aggregate: bool,
        max_depth: Optional[int],
    ) -> List[str]:

        if max_depth is not None and depth > max_depth:
            return []

        node_log = [self._print_line(depth, node, aggregate, max_depth)]

        children_logs = [
            self._get_node_log(child, depth + 1, aggregate, max_depth)
            for child in node.children
        ]

        return node_log + self._flatten(children_logs)

    def _print_line(
        self,
        depth: int,
        node: tracers.TraceNode,
        aggregate: bool,
        max_depth: Optional[int],
    ) -> str:
        indent = "  " * depth
        value = node.value

        if value is None:
            formatted_value = "{'avg': '?', 'max': '?', 'min': '?'}"

        elif aggregate:
            try:
                formatted_value = str(
                    {
                        "avg": numpy.mean(value),
                        "max": numpy.max(value),
                        "min": numpy.min(value),
                    }
                )

            except TypeError:
                formatted_value = "{'avg': '?', 'max': '?', 'min': '?'}"

        else:
            formatted_value = self.display(value)

        return f"{indent}{node.name}<{node.period}, ({node.branch_name})> = {formatted_value}"

    def _flatten(
        self,
        lists: List[List[str]],
    ) -> List[str]:
        return [item for list_ in lists for item in list_]



from policyengine_core.parameters import ParameterNode


class ParameterScaleBracket(ParameterNode):
    """
    A parameter scale bracket.
    """

    _allowed_keys = set(
        ["amount", "threshold", "rate", "average_rate", "base"]
    )



import copy
import os
from typing import Dict, List, Optional

from policyengine_core.errors import ParameterParsingError
from .at_instant_like import AtInstantLike
from .parameter_at_instant import ParameterAtInstant

from .helpers import _validate_parameter, _compose_name
from .config import COMMON_KEYS
from policyengine_core.commons.misc import empty_clone
from policyengine_core.periods import INSTANT_PATTERN, period as get_period


class Parameter(AtInstantLike):
    """A parameter of the legislation.

    Parameters can change over time.

    Attributes:
        values_list: List of the values, in reverse chronological order.


    Args:
        name: Name of the parameter, e.g. "taxes.some_tax.some_param".
        data: Data loaded from a YAML file.
        file_path: File the parameter was loaded from.

    Instantiate a parameter without metadata:

    >>>  Parameter('rate', data = {
            "2015-01-01": 550,
            "2016-01-01": 600
            })

    Instantiate a parameter with metadata:

    >>>  Parameter('rate', data = {
            'description': 'Income tax rate applied on salaries',
            'values': {
                "2015-01-01": {'value': 550, 'metadata': {'reference': 'http://taxes.gov/income_tax/2015'}},
                "2016-01-01": {'value': 600, 'metadata': {'reference': 'http://taxes.gov/income_tax/2016'}}
                }
            })

    """

    def __init__(
        self, name: str, data: dict, file_path: Optional[str] = None
    ) -> None:
        self.name: str = name
        self.file_path: Optional[str] = file_path
        _validate_parameter(self, data, data_type=dict)
        self.description: Optional[str] = None
        self.metadata: Dict = {}
        self.documentation: Optional[str] = None

        # Normal parameter declaration: the values are declared under the 'values' key: parse the description and metadata.
        if data.get("values"):
            # 'unit' and 'reference' are only listed here for backward compatibility
            self.metadata.update(data.get("metadata", {}))
            _validate_parameter(
                self, data, allowed_keys=COMMON_KEYS.union({"values"})
            )
            self.description = data.get("description")

            _validate_parameter(self, data["values"], data_type=dict)
            values = data["values"]

            self.documentation = data.get("documentation")

        else:  # Simplified parameter declaration: only values are provided
            values = data

        instants = sorted(
            values.keys(), reverse=True
        )  # sort in reverse chronological order

        values_list = []
        for instant_str in instants:
            if not INSTANT_PATTERN.match(instant_str):
                raise ParameterParsingError(
                    "Invalid property '{}' in '{}'. Properties must be valid YYYY-MM-DD instants, such as 2017-01-15.".format(
                        instant_str, self.name
                    ),
                    file_path,
                )

            instant_info = values[instant_str]

            #  Ignore expected values, as they are just metadata
            if (
                instant_info == "expected"
                or isinstance(instant_info, dict)
                and instant_info.get("expected")
            ):
                continue

            value_name = _compose_name(name, item_name=instant_str)
            value_at_instant = ParameterAtInstant(
                value_name,
                instant_str,
                data=instant_info,
                file_path=self.file_path,
                metadata=self.metadata,
            )
            values_list.append(value_at_instant)

        self.values_list: List[ParameterAtInstant] = values_list

    def __repr__(self):
        return os.linesep.join(
            [
                "{}: {}".format(
                    value.instant_str,
                    value.value if value.value is not None else "null",
                )
                for value in self.values_list
            ]
        )

    def __eq__(self, other):
        return (self.name == other.name) and (
            self.values_list == other.values_list
        )

    def clone(self):
        clone = empty_clone(self)
        clone.__dict__ = self.__dict__.copy()

        clone.metadata = copy.deepcopy(self.metadata)
        clone.values_list = [
            parameter_at_instant.clone()
            for parameter_at_instant in self.values_list
        ]
        return clone

    def update(self, period=None, start=None, stop=None, value=None):
        """
        Change the value for a given period.

        :param period: Period where the value is modified. If set, `start` and `stop` should be `None`.
        :param start: Start of the period. Instance of `policyengine_core.Instant`. If set, `period` should be `None`.
        :param stop: Stop of the period. Instance of `policyengine_core.Instant`. If set, `period` should be `None`.
        :param value: New value. If `None`, the parameter is removed from the legislation parameters for the given period.
        """
        if period is not None:
            if start is not None or stop is not None:
                raise TypeError(
                    "Wrong input for 'update' method: use either 'update(period, value = value)' or 'update(start = start, stop = stop, value = value)'. You cannot both use 'period' and 'start' or 'stop'."
                )
            if isinstance(period, str):
                period = get_period(period)
            start = period.start
            stop = period.stop
        if start is None:
            raise ValueError("You must provide either a start or a period")
        start_str = str(start)
        stop_str = str(stop.offset(1, "day")) if stop else None

        old_values = self.values_list
        new_values = []
        n = len(old_values)
        i = 0

        # Future intervals : not affected
        if stop_str:
            while (i < n) and (old_values[i].instant_str >= stop_str):
                new_values.append(old_values[i])
                i += 1

        # Right-overlapped interval
        if stop_str:
            if new_values and (stop_str == new_values[-1].instant_str):
                pass  # such interval is empty
            else:
                if i < n:
                    overlapped_value = old_values[i].value
                    value_name = _compose_name(self.name, item_name=stop_str)
                    new_interval = ParameterAtInstant(
                        value_name, stop_str, data={"value": overlapped_value}
                    )
                    new_values.append(new_interval)
                else:
                    value_name = _compose_name(self.name, item_name=stop_str)
                    new_interval = ParameterAtInstant(
                        value_name, stop_str, data={"value": None}
                    )
                    new_values.append(new_interval)

        # Insert new interval
        value_name = _compose_name(self.name, item_name=start_str)
        new_interval = ParameterAtInstant(
            value_name, start_str, data={"value": value}
        )
        new_values.append(new_interval)

        # Remove covered intervals
        while (i < n) and (old_values[i].instant_str >= start_str):
            i += 1

        # Past intervals : not affected
        while i < n:
            new_values.append(old_values[i])
            i += 1

        self.values_list = new_values

    def get_descendants(self):
        return iter(())

    def _get_at_instant(self, instant):
        for value_at_instant in self.values_list:
            if value_at_instant.instant_str <= instant:
                return value_at_instant.value
        return None



import datetime
import os

from policyengine_core import periods
from policyengine_core.periods import config


def N_(message):
    return message


def instant(instant):
    """Return a new instant, aka a triple of integers (year, month, day).

    >>> instant(2014)
    Instant((2014, 1, 1))
    >>> instant('2014')
    Instant((2014, 1, 1))
    >>> instant('2014-02')
    Instant((2014, 2, 1))
    >>> instant('2014-3-2')
    Instant((2014, 3, 2))
    >>> instant(instant('2014-3-2'))
    Instant((2014, 3, 2))
    >>> instant(period('month', '2014-3-2'))
    Instant((2014, 3, 2))

    >>> instant(None)
    """
    if instant is None:
        return None
    if isinstance(instant, periods.Instant):
        return instant
    if isinstance(instant, str):
        if not config.INSTANT_PATTERN.match(instant):
            raise ValueError(
                "'{}' is not a valid instant. Instants are described using the 'YYYY-MM-DD' format, for instance '2015-06-15'.".format(
                    instant
                )
            )
        instant = periods.Instant(
            int(fragment) for fragment in instant.split("-", 2)[:3]
        )
    elif isinstance(instant, datetime.date):
        instant = periods.Instant((instant.year, instant.month, instant.day))
    elif isinstance(instant, int):
        instant = (instant,)
    elif isinstance(instant, list):
        assert 1 <= len(instant) <= 3
        instant = tuple(instant)
    elif isinstance(instant, periods.Period):
        instant = instant.start
    else:
        assert isinstance(instant, tuple), instant
        assert 1 <= len(instant) <= 3
    if len(instant) == 1:
        return periods.Instant((instant[0], 1, 1))
    if len(instant) == 2:
        return periods.Instant((instant[0], instant[1], 1))
    return periods.Instant(instant)


def instant_date(instant):
    if instant is None:
        return None
    instant_date = config.date_by_instant_cache.get(instant)
    if instant_date is None:
        config.date_by_instant_cache[instant] = instant_date = datetime.date(
            *instant
        )
    return instant_date


def period(value):
    """Return a new period, aka a triple (unit, start_instant, size).

    >>> period('2014')
    Period((YEAR, Instant((2014, 1, 1)), 1))
    >>> period('year:2014')
    Period((YEAR, Instant((2014, 1, 1)), 1))

    >>> period('2014-2')
    Period((MONTH, Instant((2014, 2, 1)), 1))
    >>> period('2014-02')
    Period((MONTH, Instant((2014, 2, 1)), 1))
    >>> period('month:2014-2')
    Period((MONTH, Instant((2014, 2, 1)), 1))

    >>> period('year:2014-2')
    Period((YEAR, Instant((2014, 2, 1)), 1))
    """
    if isinstance(value, periods.Period):
        return value

    if isinstance(value, periods.Instant):
        return periods.Period((config.DAY, value, 1))

    def parse_simple_period(value):
        """
        Parses simple periods respecting the ISO format, such as 2012 or 2015-03
        """
        try:
            date = datetime.datetime.strptime(value, "%Y")
        except ValueError:
            try:
                date = datetime.datetime.strptime(value, "%Y-%m")
            except ValueError:
                try:
                    date = datetime.datetime.strptime(value, "%Y-%m-%d")
                except ValueError:
                    return None
                else:
                    return periods.Period(
                        (
                            config.DAY,
                            periods.Instant((date.year, date.month, date.day)),
                            1,
                        )
                    )
            else:
                return periods.Period(
                    (
                        config.MONTH,
                        periods.Instant((date.year, date.month, 1)),
                        1,
                    )
                )
        else:
            return periods.Period(
                (config.YEAR, periods.Instant((date.year, date.month, 1)), 1)
            )

    def raise_error(value):
        message = os.linesep.join(
            [
                "Expected a period (eg. '2017', '2017-01', '2017-01-01', ...); got: '{}'.".format(
                    value
                ),
                "Learn more about legal period formats in OpenFisca:",
                "<https://openfisca.org/doc/coding-the-legislation/35_periods.html#periods-in-simulations>.",
            ]
        )
        raise ValueError(message)

    if value == "ETERNITY" or value == config.ETERNITY:
        return periods.Period(
            ("eternity", instant(datetime.date.min), float("inf"))
        )

    # check the type
    if isinstance(value, int):
        return periods.Period((config.YEAR, periods.Instant((value, 1, 1)), 1))
    if not isinstance(value, str):
        raise_error(value)

    # try to parse as a simple period
    period = parse_simple_period(value)
    if period is not None:
        return period

    # complex period must have a ':' in their strings
    if ":" not in value:
        raise_error(value)

    components = value.split(":")

    # left-most component must be a valid unit
    unit = components[0]
    if unit not in (config.DAY, config.MONTH, config.YEAR):
        raise_error(value)

    # middle component must be a valid iso period
    base_period = parse_simple_period(components[1])
    if not base_period:
        raise_error(value)

    # period like year:2015-03 have a size of 1
    if len(components) == 2:
        size = 1
    # if provided, make sure the size is an integer
    elif len(components) == 3:
        try:
            size = int(components[2])
        except ValueError:
            raise_error(value)
    # if there is more than 2 ":" in the string, the period is invalid
    else:
        raise_error(value)

    # reject ambiguous period such as month:2014
    if unit_weight(base_period.unit) > unit_weight(unit):
        raise_error(value)

    return periods.Period((unit, base_period.start, size))


def key_period_size(period):
    """
    Defines a key in order to sort periods by length. It uses two aspects : first unit then size

    :param period: an OpenFisca period
    :return: a string

    >>> key_period_size(period('2014'))
    '2_1'
    >>> key_period_size(period('2013'))
    '2_1'
    >>> key_period_size(period('2014-01'))
    '1_1'

    """

    unit, start, size = period

    return "{}_{}".format(unit_weight(unit), size)


def unit_weights():
    return {
        config.DAY: 100,
        config.MONTH: 200,
        config.YEAR: 300,
        config.ETERNITY: 400,
    }


def unit_weight(unit):
    return unit_weights()[unit]



import copy
import typing
from typing import TYPE_CHECKING, Any, List

import dpath.util
import numpy as np
from numpy.typing import ArrayLike

from policyengine_core import periods
from policyengine_core.entities import Entity, Role
from policyengine_core.errors import (
    PeriodMismatchError,
    SituationParsingError,
    VariableNotFoundError,
)
from policyengine_core.periods import Period
from policyengine_core.populations import GroupPopulation, Population
from policyengine_core.simulations.simulation import Simulation
from policyengine_core.simulations.helpers import (
    check_type,
    _get_person_count,
    transform_to_strict_syntax,
)

if TYPE_CHECKING:
    from policyengine_core.taxbenefitsystems.tax_benefit_system import (
        TaxBenefitSystem,
    )

from policyengine_core.variables import Variable


class SimulationBuilder:
    def __init__(self):
        self.default_period: Period = None  # Simulation period used for variables when no period is defined
        self.persons_plural: str = None  # Plural name for person entity in current tax and benefits system

        # JSON input - Memory of known input values. Indexed by variable or axis name.
        self.input_buffer: typing.Dict[
            Variable.name, typing.Dict[str(periods.period), np.array]
        ] = {}
        self.populations: typing.Dict[Entity.key, Population] = {}
        # JSON input - Number of items of each entity type. Indexed by entities plural names. Should be consistent with ``entity_ids``, including axes.
        self.entity_counts: typing.Dict[Entity.plural, int] = {}
        # JSON input - typing.List of items of each entity type. Indexed by entities plural names. Should be consistent with ``entity_counts``.
        self.entity_ids: typing.Dict[Entity.plural, typing.List[int]] = {}

        # Links entities with persons. For each person index in persons ids list, set entity index in entity ids id. E.g.: self.memberships[entity.plural][person_index] = entity_ids.index(instance_id)
        self.memberships: typing.Dict[Entity.plural, typing.List[int]] = {}
        self.roles: typing.Dict[Entity.plural, typing.List[int]] = {}

        self.variable_entities: typing.Dict[Variable.name, Entity] = {}

        self.axes = [[]]
        self.axes_entity_counts: typing.Dict[Entity.plural, int] = {}
        self.axes_entity_ids: typing.Dict[Entity.plural, typing.List[int]] = {}
        self.axes_memberships: typing.Dict[
            Entity.plural, typing.List[int]
        ] = {}
        self.axes_roles: typing.Dict[Entity.plural, typing.List[int]] = {}

    def build_from_dict(
        self,
        tax_benefit_system: "TaxBenefitSystem",
        input_dict: dict,
        simulation: Simulation = None,
    ) -> Simulation:
        """
        Build a simulation from ``input_dict``. Optionally overwrites an existing simulation.

        This method uses :any:`build_from_entities` if entities are fully specified, or :any:`build_from_variables` if not.
        """

        input_dict = self.explicit_singular_entities(
            tax_benefit_system, input_dict
        )
        if any(
            key in tax_benefit_system.entities_plural()
            for key in input_dict.keys()
        ):
            return self.build_from_entities(
                tax_benefit_system, input_dict, simulation
            )
        else:
            return self.build_from_variables(
                tax_benefit_system, input_dict, simulation
            )

    def build_from_entities(
        self,
        tax_benefit_system: "TaxBenefitSystem",
        input_dict: dict,
        simulation: Simulation = None,
    ) -> Simulation:
        """
        Build a simulation from a Python dict ``input_dict`` fully specifying entities.

        Examples:

        >>> simulation_builder.build_from_entities({
            'persons': {'Javier': { 'salary': {'2018-11': 2000}}},
            'households': {'household': {'parents': ['Javier']}}
            })
        """
        input_dict = copy.deepcopy(input_dict)

        if simulation is None:
            simulation = Simulation(
                tax_benefit_system=tax_benefit_system,
                populations=tax_benefit_system.instantiate_entities(),
            )

        # Register variables so get_variable_entity can find them
        for (variable_name, _variable) in tax_benefit_system.variables.items():
            self.register_variable(
                variable_name,
                simulation.get_variable_population(variable_name).entity,
            )

        check_type(input_dict, dict, ["error"])
        axes = input_dict.pop("axes", None)

        unexpected_entities = [
            entity
            for entity in input_dict
            if entity not in tax_benefit_system.entities_plural()
        ]
        if unexpected_entities:
            unexpected_entity = unexpected_entities[0]
            raise SituationParsingError(
                [unexpected_entity],
                "".join(
                    [
                        "Some entities in the situation are not defined in the loaded tax and benefit system.",
                        "These entities are not found: {0}.",
                        "The defined entities are: {1}.",
                    ]
                ).format(
                    ", ".join(unexpected_entities),
                    ", ".join(tax_benefit_system.entities_plural()),
                ),
            )
        persons_json = input_dict.get(
            tax_benefit_system.person_entity.plural, None
        )

        if not persons_json:
            raise SituationParsingError(
                [tax_benefit_system.person_entity.plural],
                "No {0} found. At least one {0} must be defined to run a simulation.".format(
                    tax_benefit_system.person_entity.key
                ),
            )

        persons_ids = self.add_person_entity(
            simulation.persons.entity, persons_json
        )

        for entity_class in tax_benefit_system.group_entities:
            instances_json = input_dict.get(entity_class.plural)
            if instances_json is not None:
                self.add_group_entity(
                    self.persons_plural,
                    persons_ids,
                    entity_class,
                    instances_json,
                )
            else:
                self.add_default_group_entity(persons_ids, entity_class)

        if axes:
            self.axes = axes
            self.expand_axes()

        try:
            self.finalize_variables_init(simulation.persons)
        except PeriodMismatchError as e:
            self.raise_period_mismatch(
                simulation.persons.entity, persons_json, e
            )

        for entity_class in tax_benefit_system.group_entities:
            try:
                population = simulation.populations[entity_class.key]
                self.finalize_variables_init(population)
            except PeriodMismatchError as e:
                self.raise_period_mismatch(
                    population.entity, instances_json, e
                )

        return simulation

    def build_from_variables(
        self,
        tax_benefit_system: "TaxBenefitSystem",
        input_dict: dict,
        simulation: Simulation = None,
    ) -> Simulation:
        """
        Build a simulation from a Python dict ``input_dict`` describing variables values without expliciting entities.

        This method uses :any:`build_default_simulation` to infer an entity structure

        Example:

        >>> simulation_builder.build_from_variables(
            {'salary': {'2016-10': 12000}}
            )
        """
        count = _get_person_count(input_dict)
        simulation = self.build_default_simulation(
            tax_benefit_system, count, simulation
        )
        for variable, value in input_dict.items():
            if not isinstance(value, dict):
                if self.default_period is None:
                    raise SituationParsingError(
                        [variable],
                        "Can't deal with type: expected object. Input variables should be set for specific periods. For instance: {'salary': {'2017-01': 2000, '2017-02': 2500}}, or {'birth_date': {'ETERNITY': '1980-01-01'}}.",
                    )
                simulation.set_input(variable, self.default_period, value)
            else:
                for period_str, dated_value in value.items():
                    simulation.set_input(variable, period_str, dated_value)
        return simulation

    def build_default_simulation(
        self,
        tax_benefit_system: "TaxBenefitSystem",
        count: int = 1,
        simulation: Simulation = None,
    ) -> Simulation:
        """
        Build a simulation where:
            - There are ``count`` persons
            - There are ``count`` instances of each group entity, containing one person
            - Every person has, in each entity, the first role
        """

        if simulation is None:
            simulation = Simulation(
                tax_benefit_system=tax_benefit_system,
                populations=tax_benefit_system.instantiate_entities(),
            )
        for population in simulation.populations.values():
            population.count = count
            population.ids = np.array(range(count))
            if not population.entity.is_person:
                population.members_entity_id = (
                    population.ids
                )  # Each person is its own group entity
        return simulation

    def create_entities(self, tax_benefit_system: "TaxBenefitSystem") -> None:
        self.populations = tax_benefit_system.instantiate_entities()

    def declare_person_entity(
        self, person_singular: str, persons_ids: typing.Iterable
    ) -> None:
        person_instance = self.populations[person_singular]
        person_instance.ids = np.array(persons_ids)
        person_instance.count = len(person_instance.ids)

        self.persons_plural = person_instance.entity.plural

    def declare_entity(
        self, entity_singular: str, entity_ids: typing.Iterable
    ) -> Population:
        entity_instance = self.populations[entity_singular]
        entity_instance.ids = np.array(entity_ids)
        entity_instance.count = len(entity_instance.ids)
        return entity_instance

    def nb_persons(self, entity_singular: str, role: Role = None) -> ArrayLike:
        return self.populations[entity_singular].nb_persons(role=role)

    def join_with_persons(
        self,
        group_population: GroupPopulation,
        persons_group_assignment: ArrayLike,
        roles: typing.Iterable[str],
    ) -> None:
        # Maps group's identifiers to a 0-based integer range, for indexing into members_roles (see PR#876)
        group_sorted_indices = np.unique(
            persons_group_assignment, return_inverse=True
        )[1]
        group_population.members_entity_id = np.argsort(group_population.ids)[
            group_sorted_indices
        ]

        flattened_roles = group_population.entity.flattened_roles
        roles_array = np.array(roles)
        if np.issubdtype(roles_array.dtype, np.integer):
            group_population.members_role = np.array(flattened_roles)[
                roles_array
            ]
        else:
            if len(flattened_roles) == 0:
                group_population.members_role = np.int64(0)
            else:
                group_population.members_role = np.select(
                    [roles_array == role.key for role in flattened_roles],
                    flattened_roles,
                )

    def build(self, tax_benefit_system: "TaxBenefitSystem") -> Simulation:
        return Simulation(
            tax_benefit_system=tax_benefit_system, populations=self.populations
        )

    def explicit_singular_entities(
        self, tax_benefit_system: "TaxBenefitSystem", input_dict: dict
    ) -> dict:
        """
        Preprocess ``input_dict`` to explicit entities defined using the single-entity shortcut

        Example:

        >>> simulation_builder.explicit_singular_entities(
            {'persons': {'Javier': {}, }, 'household': {'parents': ['Javier']}}
            )
        >>> {'persons': {'Javier': {}}, 'households': {'household': {'parents': ['Javier']}}
        """

        axes = input_dict.get("axes")

        singular_keys = set(input_dict).intersection(
            tax_benefit_system.entities_by_singular()
        )
        if not singular_keys:
            return input_dict

        result = {
            entity_id: entity_description
            for (entity_id, entity_description) in input_dict.items()
            if entity_id in tax_benefit_system.entities_plural()
        }  # filter out the singular entities

        for singular in singular_keys:
            plural = tax_benefit_system.entities_by_singular()[singular].plural
            result[plural] = {singular: input_dict[singular]}

        if axes is not None:
            result["axes"] = axes

        return result

    def add_person_entity(
        self, entity: Entity, instances_json: dict
    ) -> List[int]:
        """
        Add the simulation's instances of the persons entity as described in ``instances_json``.
        """
        check_type(instances_json, dict, [entity.plural])
        entity_ids = list(map(str, instances_json.keys()))
        self.persons_plural = entity.plural
        self.entity_ids[self.persons_plural] = entity_ids
        self.entity_counts[self.persons_plural] = len(entity_ids)

        for instance_id, instance_object in instances_json.items():
            check_type(instance_object, dict, [entity.plural, instance_id])
            self.init_variable_values(
                entity, instance_object, str(instance_id)
            )

        return self.get_ids(entity.plural)

    def add_default_group_entity(
        self, persons_ids: ArrayLike, entity: Entity
    ) -> None:
        persons_count = len(persons_ids)
        self.entity_ids[entity.plural] = [entity.key]
        self.entity_counts[entity.plural] = 1
        self.memberships[entity.plural] = np.zeros(
            persons_count, dtype=np.int32
        )
        self.roles[entity.plural] = np.repeat(
            entity.flattened_roles[0], persons_count
        )

    def add_group_entity(
        self,
        persons_plural: str,
        persons_ids: ArrayLike,
        entity: Entity,
        instances_json: dict,
    ) -> None:
        """
        Add all instances of one of the model's entities as described in ``instances_json``.
        """
        check_type(instances_json, dict, [entity.plural])
        entity_ids = list(map(str, instances_json.keys()))

        self.entity_ids[entity.plural] = entity_ids
        self.entity_counts[entity.plural] = len(entity_ids)

        persons_count = len(persons_ids)
        persons_to_allocate = set(persons_ids)
        self.memberships[entity.plural] = np.empty(
            persons_count, dtype=np.int32
        )
        self.roles[entity.plural] = np.empty(persons_count, dtype=object)

        self.entity_ids[entity.plural] = entity_ids
        self.entity_counts[entity.plural] = len(entity_ids)

        for instance_id, instance_object in instances_json.items():
            check_type(instance_object, dict, [entity.plural, instance_id])

            variables_json = (
                instance_object.copy()
            )  # Don't mutate function input

            roles_json = {
                role.plural
                or role.key: transform_to_strict_syntax(
                    variables_json.pop(role.plural or role.key, [])
                )
                for role in entity.roles
            }

            for role_id, role_definition in roles_json.items():
                check_type(
                    role_definition,
                    list,
                    [entity.plural, instance_id, role_id],
                )
                for index, person_id in enumerate(role_definition):
                    entity_plural = entity.plural
                    self.check_persons_to_allocate(
                        persons_plural,
                        entity_plural,
                        persons_ids,
                        person_id,
                        instance_id,
                        role_id,
                        persons_to_allocate,
                        index,
                    )

                    persons_to_allocate.discard(person_id)

            entity_index = entity_ids.index(instance_id)
            role_by_plural = {
                role.plural or role.key: role for role in entity.roles
            }

            for role_plural, persons_with_role in roles_json.items():
                role = role_by_plural[role_plural]

                if role.max is not None and len(persons_with_role) > role.max:
                    raise SituationParsingError(
                        [entity.plural, instance_id, role_plural],
                        f"There can be at most {role.max} {role_plural} in a {entity.key}. {len(persons_with_role)} were declared in '{instance_id}'.",
                    )

                for index_within_role, person_id in enumerate(
                    persons_with_role
                ):
                    person_index = persons_ids.index(person_id)
                    self.memberships[entity.plural][
                        person_index
                    ] = entity_index
                    person_role = (
                        role.subroles[index_within_role]
                        if role.subroles
                        else role
                    )
                    self.roles[entity.plural][person_index] = person_role

            self.init_variable_values(entity, variables_json, instance_id)

        if persons_to_allocate:
            entity_ids = entity_ids + list(persons_to_allocate)
            for person_id in persons_to_allocate:
                person_index = persons_ids.index(person_id)
                self.memberships[entity.plural][
                    person_index
                ] = entity_ids.index(person_id)
                self.roles[entity.plural][
                    person_index
                ] = entity.flattened_roles[0]
            # Adjust previously computed ids and counts
            self.entity_ids[entity.plural] = entity_ids
            self.entity_counts[entity.plural] = len(entity_ids)

        # Convert back to Python array
        self.roles[entity.plural] = self.roles[entity.plural].tolist()
        self.memberships[entity.plural] = self.memberships[
            entity.plural
        ].tolist()

    def set_default_period(self, period_str: str) -> None:
        if period_str:
            self.default_period = str(periods.period(period_str))

    def get_input(self, variable: str, period_str: str) -> Any:
        if variable not in self.input_buffer:
            self.input_buffer[variable] = {}
        return self.input_buffer[variable].get(period_str)

    def check_persons_to_allocate(
        self,
        persons_plural,
        entity_plural,
        persons_ids,
        person_id,
        entity_id,
        role_id,
        persons_to_allocate,
        index,
    ):
        check_type(
            person_id, str, [entity_plural, entity_id, role_id, str(index)]
        )
        if person_id not in persons_ids:
            raise SituationParsingError(
                [entity_plural, entity_id, role_id],
                "Unexpected value: {0}. {0} has been declared in {1} {2}, but has not been declared in {3}.".format(
                    person_id, entity_id, role_id, persons_plural
                ),
            )
        if person_id not in persons_to_allocate:
            raise SituationParsingError(
                [entity_plural, entity_id, role_id],
                "{} has been declared more than once in {}".format(
                    person_id, entity_plural
                ),
            )

    def init_variable_values(self, entity, instance_object, instance_id):
        for variable_name, variable_values in instance_object.items():
            path_in_json = [entity.plural, instance_id, variable_name]
            try:
                entity.check_variable_defined_for_entity(variable_name)
            except ValueError as e:  # The variable is defined for another entity
                raise SituationParsingError(path_in_json, e.args[0])
            except VariableNotFoundError as e:  # The variable doesn't exist
                raise SituationParsingError(path_in_json, str(e), code=404)

            instance_index = self.get_ids(entity.plural).index(instance_id)

            if not isinstance(variable_values, dict):
                if self.default_period is None:
                    raise SituationParsingError(
                        path_in_json,
                        "Can't deal with type: expected object. Input variables should be set for specific periods. For instance: {'salary': {'2017-01': 2000, '2017-02': 2500}}, or {'birth_date': {'ETERNITY': '1980-01-01'}}.",
                    )
                variable_values = {self.default_period: variable_values}

            for period_str, value in variable_values.items():
                try:
                    periods.period(period_str)
                except ValueError as e:
                    raise SituationParsingError(path_in_json, e.args[0])
                variable = entity.get_variable(variable_name)
                self.add_variable_value(
                    entity,
                    variable,
                    instance_index,
                    instance_id,
                    period_str,
                    value,
                )

    def add_variable_value(
        self, entity, variable, instance_index, instance_id, period_str, value
    ):
        path_in_json = [entity.plural, instance_id, variable.name, period_str]

        if value is None:
            return

        array = self.get_input(variable.name, str(period_str))

        if array is None:
            array_size = self.get_count(entity.plural)
            array = variable.default_array(array_size)

        try:
            value = variable.check_set_value(value)
        except ValueError as error:
            raise SituationParsingError(path_in_json, *error.args)

        array[instance_index] = value

        self.input_buffer[variable.name][
            str(periods.period(period_str))
        ] = array

    def finalize_variables_init(self, population):
        # Due to set_input mechanism, we must bufferize all inputs, then actually set them,
        # so that the months are set first and the years last.
        plural_key = population.entity.plural
        if plural_key in self.entity_counts:
            population.count = self.get_count(plural_key)
            population.ids = self.get_ids(plural_key)
        if plural_key in self.memberships:
            population.members_entity_id = np.array(
                self.get_memberships(plural_key)
            )
            population.members_role = np.array(self.get_roles(plural_key))
        for variable_name in self.input_buffer.keys():
            try:
                holder = population.get_holder(variable_name)
            except ValueError:  # Wrong entity, we can just ignore that
                continue
            buffer = self.input_buffer[variable_name]
            unsorted_periods = [
                periods.period(period_str)
                for period_str in self.input_buffer[variable_name].keys()
            ]
            # We need to handle small periods first for set_input to work
            sorted_periods = sorted(
                unsorted_periods, key=periods.key_period_size
            )
            for period_value in sorted_periods:
                values = buffer[str(period_value)]
                # Hack to replicate the values in the persons entity
                # when we have an axis along a group entity but not persons
                array = np.tile(values, population.count // len(values))
                variable = holder.variable
                # TODO - this duplicates the check in Simulation.set_input, but
                # fixing that requires improving Simulation's handling of entities
                if (variable.end is None) or (
                    period_value.start.date <= variable.end
                ):
                    holder.set_input(period_value, array)

    def raise_period_mismatch(self, entity, json, e):
        # This error happens when we try to set a variable value for a period that doesn't match its definition period
        # It is only raised when we consume the buffer. We thus don't know which exact key caused the error.
        # We do a basic research to find the culprit path
        culprit_path = next(
            dpath.util.search(
                json,
                "*/{}/{}".format(e.variable_name, str(e.period)),
                yielded=True,
            ),
            None,
        )
        if culprit_path:
            path = [entity.plural] + culprit_path[0].split("/")
        else:
            path = [
                entity.plural
            ]  # Fallback: if we can't find the culprit, just set the error at the entities level

        raise SituationParsingError(path, e.message)

    # Returns the total number of instances of this entity, including when there is replication along axes
    def get_count(self, entity_name):
        return self.axes_entity_counts.get(
            entity_name, self.entity_counts[entity_name]
        )

    # Returns the ids of instances of this entity, including when there is replication along axes
    def get_ids(self, entity_name):
        return self.axes_entity_ids.get(
            entity_name, self.entity_ids[entity_name]
        )

    # Returns the memberships of individuals in this entity, including when there is replication along axes
    def get_memberships(self, entity_name):
        # Return empty array for the "persons" entity
        return self.axes_memberships.get(
            entity_name, self.memberships.get(entity_name, [])
        )

    # Returns the roles of individuals in this entity, including when there is replication along axes
    def get_roles(self, entity_name):
        # Return empty array for the "persons" entity
        return self.axes_roles.get(
            entity_name, self.roles.get(entity_name, [])
        )

    def add_parallel_axis(self, axis):
        # All parallel axes have the same count and entity.
        # Search for a compatible axis, if none exists, error out
        self.axes[0].append(axis)

    def add_perpendicular_axis(self, axis):
        # This adds an axis perpendicular to all previous dimensions
        self.axes.append([axis])

    def expand_axes(self):
        # This method should be idempotent & allow change in axes
        perpendicular_dimensions = self.axes

        cell_count = 1
        for parallel_axes in perpendicular_dimensions:
            first_axis = parallel_axes[0]
            axis_count = first_axis["count"]
            cell_count *= axis_count

        # Scale the "prototype" situation, repeating it cell_count times
        for entity_name in self.entity_counts.keys():
            # Adjust counts
            self.axes_entity_counts[entity_name] = (
                self.get_count(entity_name) * cell_count
            )
            # Adjust ids
            original_ids = self.get_ids(entity_name) * cell_count
            indices = np.arange(
                0, cell_count * self.entity_counts[entity_name]
            )
            adjusted_ids = [
                id + str(ix) for id, ix in zip(original_ids, indices)
            ]
            self.axes_entity_ids[entity_name] = adjusted_ids
            # Adjust roles
            original_roles = self.get_roles(entity_name)
            adjusted_roles = np.repeat(original_roles, cell_count)
            self.axes_roles[entity_name] = adjusted_roles
            # Adjust memberships, for group entities only
            if entity_name != self.persons_plural:
                original_memberships = self.get_memberships(entity_name)
                repeated_memberships = np.repeat(
                    original_memberships, cell_count
                )
                indices = (
                    np.repeat(
                        np.arange(0, cell_count), len(original_memberships)
                    )
                    * self.entity_counts[entity_name]
                )
                adjusted_memberships = (
                    np.array(repeated_memberships) + indices
                ).tolist()
                self.axes_memberships[entity_name] = adjusted_memberships

        # Now generate input values along the specified axes
        # TODO - factor out the common logic here
        if len(self.axes) == 1 and len(self.axes[0]):
            parallel_axes = self.axes[0]
            first_axis = parallel_axes[0]
            axis_count: int = first_axis["count"]
            axis_entity = self.get_variable_entity(first_axis["name"])
            axis_entity_step_size = self.entity_counts[axis_entity.plural]
            # Distribute values along axes
            for axis in parallel_axes:
                axis_index = axis.get("index", 0)
                axis_period = axis.get("period", self.default_period)
                axis_name = axis["name"]
                variable = axis_entity.get_variable(axis_name)
                array = self.get_input(axis_name, str(axis_period))
                if array is None:
                    array = variable.default_array(
                        axis_count * axis_entity_step_size
                    )
                elif array.size == axis_entity_step_size:
                    array = np.tile(array, axis_count)
                array[axis_index::axis_entity_step_size] = np.linspace(
                    axis["min"],
                    axis["max"],
                    num=axis_count,
                )
                # Set input
                self.input_buffer[axis_name][str(axis_period)] = array
        else:
            first_axes_count: typing.List[int] = (
                parallel_axes[0]["count"] for parallel_axes in self.axes
            )
            axes_linspaces = [
                np.linspace(0, axis_count - 1, num=axis_count)
                for axis_count in first_axes_count
            ]
            axes_meshes = np.meshgrid(*axes_linspaces)
            for parallel_axes, mesh in zip(self.axes, axes_meshes):
                first_axis = parallel_axes[0]
                axis_count = first_axis["count"]
                axis_entity = self.get_variable_entity(first_axis["name"])
                axis_entity_step_size = self.entity_counts[axis_entity.plural]
                # Distribute values along the grid
                for axis in parallel_axes:
                    axis_index = axis.get("index", 0)
                    axis_period = axis.get("period") or self.default_period
                    axis_name = axis["name"]
                    variable = axis_entity.get_variable(axis_name)
                    array = self.get_input(axis_name, str(axis_period))
                    if array is None:
                        array = variable.default_array(
                            cell_count * axis_entity_step_size
                        )
                    elif array.size == axis_entity_step_size:
                        array = np.tile(array, cell_count)
                    array[axis_index::axis_entity_step_size] = axis[
                        "min"
                    ] + mesh.reshape(cell_count) * (
                        axis["max"] - axis["min"]
                    ) / (
                        axis_count - 1
                    )
                    self.input_buffer[axis_name][str(axis_period)] = array

    def get_variable_entity(self, variable_name):
        return self.variable_entities[variable_name]

    def register_variable(self, variable_name, entity):
        self.variable_entities[variable_name] = entity



"""
This file defines variables for the modelled legislation.

A variable is a property of an Entity such as a Person, a Household…

See https://openfisca.org/doc/key-concepts/variables.html
"""

# Import the Entities specifically defined for this tax and benefit system
from policyengine_core.country_template.entities import Household, Person

# Import from openfisca-core the Python objects used to code the legislation in OpenFisca
from policyengine_core.periods import MONTH
from policyengine_core.variables import Variable


class basic_income(Variable):
    value_type = float
    entity = Person
    definition_period = MONTH
    label = "Basic income provided to adults"
    reference = "https://law.gov.example/basic_income"  # Always use the most official source

    def formula_2016_12(person, period, parameters):
        """
        Basic income provided to adults.

        Since Dec 1st 2016, the basic income is provided to any adult, without considering their income.
        """
        age_condition = (
            person("age", period) >= parameters(period).general.age_of_majority
        )
        return (
            age_condition * parameters(period).benefits.basic_income
        )  # This '*' is a vectorial 'if'. See https://openfisca.org/doc/coding-the-legislation/25_vectorial_computing.html#control-structures

    def formula_2015_12(person, period, parameters):
        """
        Basic income provided to adults.

        From Dec 1st 2015 to Nov 30 2016, the basic income is provided to adults who have no income.
        Before Dec 1st 2015, the basic income does not exist in the law, and calculating it returns its default value, which is 0.
        """
        age_condition = (
            person("age", period) >= parameters(period).general.age_of_majority
        )
        salary_condition = person("salary", period) == 0
        return (
            age_condition
            * salary_condition
            * parameters(period).benefits.basic_income
        )  # The '*' is also used as a vectorial 'and'. See https://openfisca.org/doc/coding-the-legislation/25_vectorial_computing.html#boolean-operations


class housing_allowance(Variable):
    value_type = float
    entity = Household
    definition_period = MONTH
    label = "Housing allowance"
    reference = "https://law.gov.example/housing_allowance"  # Always use the most official source
    end = "2016-11-30"  # This allowance was removed on the 1st of Dec 2016. Calculating it before this date will always return the variable default value, 0.
    unit = "currency-EUR"
    documentation = """
    This allowance was introduced on the 1st of Jan 1980.
    It disappeared in Dec 2016.
    """

    def formula_1980(household, period, parameters):
        """
        Housing allowance.

        This allowance was introduced on the 1st of Jan 1980.
        Calculating it before this date will always return the variable default value, 0.

        To compute this allowance, the 'rent' value must be provided for the same month,
        but 'housing_occupancy_status' is not necessary.
        """
        return (
            household("rent", period)
            * parameters(period).benefits.housing_allowance
        )


# By default, you can use utf-8 characters in a variable. OpenFisca web API manages utf-8 encoding.
class pension(Variable):
    value_type = float
    entity = Person
    definition_period = MONTH
    label = "Pension for the elderly. Pension attribuée aux personnes âgées. تقاعد."
    reference = [
        "https://fr.wikipedia.org/wiki/Retraite_(économie)",
        "https://ar.wikipedia.org/wiki/تقاعد",
    ]

    def formula(person, period, parameters):
        """
        Pension for the elderly.

        A person's pension depends on their birth date.
        In French: retraite selon l'âge.
        In Arabic: تقاعد.
        """
        age_condition = (
            person("age", period)
            >= parameters(period).general.age_of_retirement
        )
        return age_condition


class parenting_allowance(Variable):
    value_type = float
    entity = Household
    definition_period = MONTH
    label = "Allowance for low income people with children to care for."
    documentation = "Loosely based on the Australian parenting pension."
    reference = "https://www.servicesaustralia.gov.au/individuals/services/centrelink/parenting-payment/who-can-get-it"

    def formula(household, period, parameters):
        """
        Parenting allowance for households.

        A person's parenting allowance depends on how many dependents they have,
        how much they, and their partner, earn
        if they are single with a child under 8
        or if they are partnered with a child under 6.
        """
        parenting_allowance = parameters(period).benefits.parenting_allowance

        household_income = household("household_income", period)
        income_threshold = parenting_allowance.income_threshold
        income_condition = household_income <= income_threshold

        is_single = household.nb_persons(Household.PARENT) == 1
        ages = household.members("age", period)
        under_8 = household.any(ages < 8)
        under_6 = household.any(ages < 6)

        allowance_condition = income_condition * (
            (is_single * under_8) + under_6
        )
        allowance_amount = parenting_allowance.amount

        return allowance_condition * allowance_amount


class household_income(Variable):
    value_type = float
    entity = Household
    definition_period = MONTH
    label = "The sum of the salaries of those living in a household"

    def formula(household, period, _parameters):
        """A household's income."""
        salaries = household.members("salary", period)
        return household.sum(salaries)



from policyengine_core import projectors


def projectable(function):
    """
    Decorator to indicate that when called on a projector, the outcome of the function must be projected.
    For instance person.household.sum(...) must be projected on person, while it would not make sense for person.household.get_holder.
    """
    function.projectable = True
    return function


def get_projector_from_shortcut(population, shortcut, parent=None):
    if population.entity.is_person:
        if shortcut in population.simulation.populations:
            entity_2 = population.simulation.populations[shortcut]
            return projectors.EntityToPersonProjector(entity_2, parent)
    else:
        if shortcut == "first_person":
            return projectors.FirstPersonToEntityProjector(population, parent)
        role = next(
            (
                role
                for role in population.entity.flattened_roles
                if (role.max == 1) and (role.key == shortcut)
            ),
            None,
        )
        if role:
            return projectors.UniqueRoleToEntityProjector(
                population, role, parent
            )
        if shortcut in population.entity.containing_entities:
            return getattr(
                projectors.FirstPersonToEntityProjector(population, parent),
                shortcut,
            )



import abc
import bisect
import os
import typing

from policyengine_core import tools
from policyengine_core.taxscales.tax_scale_like import TaxScaleLike


class AmountTaxScaleLike(TaxScaleLike, abc.ABC):
    """
    Base class for various types of amount-based tax scales: single amount,
    marginal amount...
    """

    amounts: typing.List

    def __init__(
        self,
        name: typing.Optional[str] = None,
        option: typing.Any = None,
        unit: typing.Any = None,
    ) -> None:
        super().__init__(name, option, unit)
        self.amounts = []

    def __repr__(self) -> str:
        return tools.indent(
            os.linesep.join(
                [
                    f"- threshold: {threshold}{os.linesep}  amount: {amount}"
                    for (threshold, amount) in zip(
                        self.thresholds, self.amounts
                    )
                ]
            )
        )

    def add_bracket(
        self,
        threshold: int,
        amount: typing.Union[int, float],
    ) -> None:
        if threshold in self.thresholds:
            i = self.thresholds.index(threshold)
            self.amounts[i] += amount

        else:
            i = bisect.bisect_left(self.thresholds, threshold)
            self.thresholds.insert(i, threshold)
            self.amounts.insert(i, amount)

    def to_dict(self) -> dict:
        return {
            str(threshold): self.amounts[index]
            for index, threshold in enumerate(self.thresholds)
        }



import copy
import typing

from policyengine_core import commons
from policyengine_core.errors import ParameterParsingError
from policyengine_core.parameters.config import ALLOWED_PARAM_TYPES
from policyengine_core.parameters.helpers import _validate_parameter


class ParameterAtInstant:
    """
    A value of a parameter at a given instant.
    """

    _allowed_keys = set(["value", "metadata"])

    def __init__(
        self,
        name: str,
        instant_str: str,
        data: dict = None,
        file_path: str = None,
        metadata: dict = None,
    ):
        """
        :param str name: name of the parameter, e.g. "taxes.some_tax.some_param"
        :param str instant_str: Date of the value in the format `YYYY-MM-DD`.
        :param dict data: Data, usually loaded from a YAML file.
        """
        self.name: str = name
        self.instant_str: str = instant_str
        self.file_path: str = file_path
        self.metadata: typing.Dict = {}

        # Accept { 2015-01-01: 4000 }
        if not isinstance(data, dict) and isinstance(
            data, ALLOWED_PARAM_TYPES
        ):
            self.value = data
            return

        if metadata is not None:
            self.metadata.update(metadata)  # Inherit metadata from Parameter
        self.metadata.update(data.get("metadata", {}))
        self.validate(data)
        self.value: float = data["value"]

    def validate(self, data: dict) -> None:
        _validate_parameter(
            self, data, data_type=dict, allowed_keys=self._allowed_keys
        )
        try:
            value = data["value"]
        except KeyError:
            raise ParameterParsingError(
                "Missing 'value' property for {}".format(self.name),
                self.file_path,
            )
        if not isinstance(value, ALLOWED_PARAM_TYPES):
            raise ParameterParsingError(
                "Value in {} has type {}, which is not one of the allowed types ({}): {}".format(
                    self.name, type(value), ALLOWED_PARAM_TYPES, value
                ),
                self.file_path,
            )

    def __eq__(self, other) -> bool:
        return (
            (self.name == other.name)
            and (self.instant_str == other.instant_str)
            and (self.value == other.value)
        )

    def __repr__(self) -> str:
        return "ParameterAtInstant({})".format({self.instant_str: self.value})

    def clone(self) -> "ParameterAtInstant":
        clone = commons.empty_clone(self)
        clone.__dict__ = self.__dict__.copy()
        clone.metadata = copy.deepcopy(self.metadata)
        return clone



from .formatting import (
    DARK_BLUE,
    BLUE,
    LIGHT_BLUE,
    LIGHTER_BLUE,
    DARK_GREEN,
    LIGHT_GREEN,
    DARK_GRAY,
    GRAY,
    LIGHT_GRAY,
    WHITE,
    format_fig,
    display_fig,
)



import os


class ParameterParsingError(Exception):
    """
    Exception raised when a parameter cannot be parsed.
    """

    def __init__(self, message, file=None, traceback=None):
        """
        :param message: Error message
        :param file: Parameter file which caused the error (optional)
        :param traceback: Traceback (optional)
        """
        if file is not None:
            message = os.linesep.join(
                ["Error parsing parameter file '{}':".format(file), message]
            )
        if traceback is not None:
            message = os.linesep.join([traceback, message])
        super(ParameterParsingError, self).__init__(message)



"""
This file defines a reform to add a dynamic variable, based on input data.

A reform is a set of modifications to be applied to a reference tax and benefit system to carry out experiments.

See https://openfisca.org/doc/key-concepts/reforms.html
"""

# Import the Entities specifically defined for this tax and benefit system
from policyengine_core.country_template.entities import Person

# Import from openfisca-core the Python objects used to code the legislation in OpenFisca
from policyengine_core.periods import MONTH
from policyengine_core.reforms import Reform
from policyengine_core.variables import Variable


def create_dynamic_variable(name, **variable):
    """Create new variable dynamically."""
    NewVariable = type(
        name,
        (Variable,),
        {
            "value_type": variable["value_type"],
            "entity": variable["entity"],
            "default_value": variable["default_value"],
            "definition_period": variable["definition_period"],
            "label": variable["label"],
            "reference": variable["reference"],
        },
    )

    return NewVariable


class add_dynamic_variable(Reform):
    def apply(self):
        """
        Apply reform.

        A reform always defines an `apply` method that builds the reformed tax
        and benefit system from the reference one.

        See https://openfisca.org/doc/coding-the-legislation/reforms.html#writing-a-reform
        """
        NewVariable = create_dynamic_variable(
            name="goes_to_school",
            value_type=bool,
            entity=Person,
            default_value=True,
            definition_period=MONTH,
            label="The person goes to school (only relevant for children)",
            reference="https://law.gov.example/goes_to_school",
        )

        self.add_variable(NewVariable)



"""
This file defines a reform.

A reform is a set of modifications to be applied to a reference tax and benefit system to carry out experiments.

See https://openfisca.org/doc/key-concepts/reforms.html
"""

# Import the Entities specifically defined for this tax and benefit system
from policyengine_core.country_template.entities import Person

# Import from openfisca-core the Python objects used to code the legislation in OpenFisca
from policyengine_core.periods import MONTH
from policyengine_core.reforms import Reform
from policyengine_core.variables import Variable


class has_car(Variable):
    value_type = bool
    entity = Person
    default_value = True
    definition_period = MONTH
    label = "The person has a car"
    reference = "https://law.gov.example/new_tax"  # Always use the most official source


class new_tax(Variable):
    value_type = float
    entity = Person
    definition_period = MONTH
    label = "New tax"
    reference = "https://law.gov.example/new_tax"  # Always use the most official source

    def formula(person, period, _parameters):
        """
        New tax reform.

        Our reform adds a new variable `new_tax` that is calculated based on
        the current `income_tax`, if the person has a car.
        """
        income_tax = person("income_tax", period)
        has_car = person("has_car", period)

        return (income_tax + 100.0) * has_car


class add_new_tax(Reform):
    def apply(self):
        """
        Apply reform.

        A reform always defines an `apply` method that builds the reformed tax
        and benefit system from the reference one.

        See https://openfisca.org/doc/coding-the-legislation/reforms.html#writing-a-reform
        """
        self.add_variable(has_car)
        self.add_variable(new_tax)



import tempfile
from typing import TYPE_CHECKING, Any, Dict, List, Type

import numpy
import numpy as np
import pandas as pd
from numpy.typing import ArrayLike

from policyengine_core import commons, periods
from policyengine_core.data.dataset import Dataset
from policyengine_core.entities.entity import Entity
from policyengine_core.enums import Enum, EnumArray
from policyengine_core.errors import CycleError, SpiralError
from policyengine_core.holders.holder import Holder
from policyengine_core.periods import Period
from policyengine_core.periods.config import ETERNITY
from policyengine_core.periods.helpers import period
from policyengine_core.tracers import (
    FullTracer,
    SimpleTracer,
    TracingParameterNodeAtInstant,
)

if TYPE_CHECKING:
    from policyengine_core.taxbenefitsystems import TaxBenefitSystem

from policyengine_core.experimental import MemoryConfig
from policyengine_core.populations import Population
from policyengine_core.tracers import SimpleTracer
from policyengine_core.variables import Variable
from policyengine_core.reforms.reform import Reform


class Simulation:
    """
    Represents a simulation, and handles the calculation logic
    """

    default_tax_benefit_system: Type["TaxBenefitSystem"] = None
    """The default tax-benefit system class to use if none is provided."""

    default_tax_benefit_system_instance: "TaxBenefitSystem" = None
    """The default tax-benefit system instance to use if none is provided. This requires that the tax-benefit system is initialised when importing a country package. This will slow down the import, but may speed up individual simulations."""

    default_dataset: Dataset = None
    """The default dataset class to use if none is provided."""

    default_dataset_year: int = None
    """The default dataset year to use if none is provided."""

    default_role: str = None
    """The default role to assign people to groups if none is provided."""

    default_input_period: str = None
    """The default period to use when inputting variables."""

    default_calculation_period: str = None
    """The default period to calculate for if none is provided."""

    def __init__(
        self,
        tax_benefit_system: "TaxBenefitSystem" = None,
        populations: Dict[str, Population] = None,
        situation: dict = None,
        dataset: Type[Dataset] = None,
        dataset_year: int = None,
        reform: Reform = None,
    ):
        if tax_benefit_system is None:
            if self.default_tax_benefit_system_instance is not None:
                tax_benefit_system = self.default_tax_benefit_system_instance
                if reform is not None:
                    tax_benefit_system = tax_benefit_system.clone()
            else:
                tax_benefit_system = self.default_tax_benefit_system()
            if reform is not None:
                reform.apply(tax_benefit_system)

        self.reform = reform
        self.tax_benefit_system = tax_benefit_system

        self.branch_name = "default"

        if dataset is None:
            if self.default_dataset is not None:
                dataset = self.default_dataset

        if dataset_year is None:
            if self.default_dataset_year is not None:
                dataset_year = self.default_dataset_year

        self.invalidated_caches = set()
        self.debug: bool = False
        self.trace: bool = False
        self.tracer: SimpleTracer = SimpleTracer()
        self.opt_out_cache: bool = False
        # controls the spirals detection; check for performance impact if > 1
        self.max_spiral_loops: int = 1
        self.memory_config: MemoryConfig = None
        self._data_storage_dir: str = None

        self.branches: Dict[str, Simulation] = {}

        if situation is not None:
            if dataset is not None:
                raise ValueError(
                    "You provided both a situation and a dataset. Only one input method is allowed."
                )
            self.build_from_populations(
                self.tax_benefit_system.instantiate_entities()
            )
            from policyengine_core.simulations.simulation_builder import (
                SimulationBuilder,
            )  # Import here to avoid circular dependency

            builder = SimulationBuilder()
            builder.default_period = self.default_input_period
            builder.build_from_dict(self.tax_benefit_system, situation, self)

        if populations is not None:
            self.build_from_populations(populations)

        if dataset is not None:
            self.dataset = dataset
            self.dataset_year = dataset_year
            self.build_from_dataset()

        # Backwards compatibility methods
        self.calc = self.calculate
        self.df = self.calculate_dataframe

    def build_from_populations(
        self, populations: Dict[str, Population]
    ) -> None:
        """This method of initialisation requires the populations to be pre-initialised.

        Args:
            populations (Dict[str, Population]): A dictionary of populations, indexed by entity key.
        """
        self.populations = populations
        self.link_to_entities_instances()
        self.create_shortcuts()

        self.populations = populations
        self.persons: Population = self.populations[
            self.tax_benefit_system.person_entity.key
        ]
        self.link_to_entities_instances()
        self.create_shortcuts()

    def build_from_dataset(self) -> None:
        """Build a simulation from a dataset."""
        self.build_from_populations(
            self.tax_benefit_system.instantiate_entities()
        )
        from policyengine_core.simulations.simulation_builder import (
            SimulationBuilder,
        )  # Import here to avoid circular dependency

        builder = SimulationBuilder()
        builder.populations = self.populations
        try:
            data = self.dataset.load(self.dataset_year)
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"The dataset file {self.dataset.name} (with year {self.dataset_year}) could not be found. "
                + "Make sure you have downloaded or built it using the `policyengine-core data` command."
            ) from e

        person_entity = self.tax_benefit_system.person_entity
        entity_id_field = f"{person_entity.key}_id"
        assert (
            entity_id_field in data
        ), f"Missing {entity_id_field} column in the dataset. Each person entity must have an ID array defined for ETERNITY."

        get_eternity_array = (
            lambda ds: ds[ETERNITY]
            if self.dataset.data_format == Dataset.TIME_PERIOD_ARRAYS
            else ds
        )
        entity_ids = get_eternity_array(data[entity_id_field])
        builder.declare_person_entity(person_entity.key, entity_ids)

        for group_entity in self.tax_benefit_system.group_entities:
            entity_id_field = f"{group_entity.key}_id"
            assert (
                entity_id_field in data
            ), f"Missing {entity_id_field} column in the dataset. Each group entity must have an ID array defined for ETERNITY."

            entity_ids = get_eternity_array(data[entity_id_field])
            builder.declare_entity(group_entity.key, entity_ids)

            person_membership_id_field = (
                f"{person_entity.key}_{group_entity.key}_id"
            )
            assert (
                person_membership_id_field in data
            ), f"Missing {person_membership_id_field} column in the dataset. Each group entity must have a person membership array defined for ETERNITY."
            person_membership_ids = get_eternity_array(
                data[person_membership_id_field]
            )

            person_role_field = f"{person_entity.key}_{group_entity.key}_role"
            if person_role_field in data:
                person_roles = get_eternity_array(data[person_role_field])
            elif "role" in data:
                person_roles = get_eternity_array(data["role"])
            elif self.default_role is not None:
                person_roles = np.full(len(entity_ids), self.default_role)
            else:
                raise ValueError(
                    f"Missing {person_role_field} column in the dataset. Each group entity must have a person role array defined for ETERNITY."
                )
            builder.join_with_persons(
                self.populations[group_entity.key],
                person_membership_ids,
                person_roles,
            )

        self.build_from_populations(builder.populations)

        for variable in data:
            if variable in self.tax_benefit_system.variables:
                if self.dataset.data_format == Dataset.TIME_PERIOD_ARRAYS:
                    for time_period in data[variable]:
                        self.set_input(
                            variable, time_period, data[variable][time_period]
                        )
                else:
                    self.set_input(variable, self.dataset_year, data[variable])
            else:
                # Silently skip.
                pass

    @property
    def trace(self) -> bool:
        return self._trace

    @trace.setter
    def trace(self, trace: SimpleTracer) -> None:
        self._trace = trace
        if trace:
            self.tracer = FullTracer()
        else:
            self.tracer = SimpleTracer()

    def link_to_entities_instances(self) -> None:
        for _key, entity_instance in self.populations.items():
            entity_instance.simulation = self

    def create_shortcuts(self) -> None:
        for _key, population in self.populations.items():
            # create shortcut simulation.person and simulation.household (for instance)
            setattr(self, population.entity.key, population)

    @property
    def data_storage_dir(self) -> str:
        """
        Temporary folder used to store intermediate calculation data in case the memory is saturated
        """
        if self._data_storage_dir is None:
            self._data_storage_dir = tempfile.mkdtemp(prefix="openfisca_")
            message = [
                (
                    "Intermediate results will be stored on disk in {} in case of memory overflow."
                ).format(self._data_storage_dir),
                "You should remove this directory once you're done with your simulation.",
            ]
        return self._data_storage_dir

    # ----- Calculation methods ----- #

    def calculate(
        self,
        variable_name: str,
        period: Period = None,
        map_to: str = None,
    ) -> ArrayLike:
        """Calculate ``variable_name`` for ``period``.

        Args:
            variable_name (str): The name of the variable to calculate.
            period (Period): The period to calculate the variable for.
            map_to (str): The name of the variable to map the result to. If None, the result is returned as is.

        Returns:
            ArrayLike: The calculated variable.
        """

        if period is not None and not isinstance(period, Period):
            period = periods.period(period)
        elif period is None and self.default_calculation_period is not None:
            period = periods.period(self.default_calculation_period)

        self.tracer.record_calculation_start(
            variable_name, period, self.branch_name
        )

        try:
            result = self._calculate(variable_name, period)
            self.tracer.record_calculation_result(result)
            if map_to is not None:
                source_entity = self.tax_benefit_system.get_variable(
                    variable_name
                ).entity.key
                result = self.map_result(result, source_entity, map_to)
            return result
        finally:
            self.tracer.record_calculation_end()
            self.purge_cache_of_invalid_values()

    def map_result(
        self,
        values: ArrayLike,
        source_entity: str,
        target_entity: str,
        how: str = None,
    ):
        """Maps values from one entity to another.

        Args:
            arr (np.array): The values in their original position.
            source_entity (str): The source entity key.
            target_entity (str): The target entity key.
            how (str, optional): A function to use when mapping. Defaults to None.

        Raises:
            ValueError: If an invalid (dis)aggregation function is passed.

        Returns:
            np.array: The mapped values.
        """
        entity_pop = self.populations[source_entity]
        target_pop = self.populations[target_entity]
        if (
            source_entity == "person"
            and target_entity in self.tax_benefit_system.group_entity_keys
        ):
            if how and how not in (
                "sum",
                "any",
                "min",
                "max",
                "all",
                "value_from_first_person",
            ):
                raise ValueError("Not a valid function.")
            return target_pop.__getattribute__(how or "sum")(values)
        elif (
            source_entity in self.tax_benefit_system.group_entity_keys
            and target_entity == "person"
        ):
            if not how:
                return entity_pop.project(values)
            if how == "mean":
                return entity_pop.project(values / entity_pop.nb_persons())
        elif source_entity == target_entity:
            return values
        else:
            return self.map_result(
                self.map_result(
                    values,
                    source_entity,
                    self.tax_benefit_system.person_entity.key,
                    how="mean",
                ),
                "person",
                target_entity,
                how="sum",
            )

    def calculate_dataframe(
        self,
        variable_names: List[str],
        period: Period = None,
        map_to: str = None,
    ) -> pd.DataFrame:
        """Calculate ``variable_names`` for ``period``.

        Args:
            variable_names (List[str]): A list of variable names to calculate.
            period (Period): The period to calculate for.

        Returns:
            pd.DataFrame: A dataframe containing the calculated variables.
        """

        df = pd.DataFrame()
        entities = [
            self.tax_benefit_system.get_variable(variable_name).entity.key
            for variable_name in variable_names
        ]
        # Check that all variables are from the same entity. If not, map values to the entity of the first variable.
        entity = entities[0]
        if not all(entity == e for e in entities):
            map_to = entity
        for variable_name in variable_names:
            df[variable_name] = self.calculate(variable_name, period, map_to)
        return df

    def _calculate(
        self, variable_name: str, period: Period = None
    ) -> ArrayLike:
        """
        Calculate the variable ``variable_name`` for the period ``period``, using the variable formula if it exists.

        Args:
            variable_name (str): The name of the variable to calculate.
            period (Period): The period to calculate the variable for.

        Returns:
            ArrayLike: The calculated variable.
        """
        population = self.get_variable_population(variable_name)
        holder = population.get_holder(variable_name)
        variable = self.tax_benefit_system.get_variable(
            variable_name, check_existence=True
        )

        self._check_period_consistency(period, variable)

        # First look for a value already cached
        cached_array = holder.get_array(period)
        if cached_array is not None:
            return cached_array

        if variable.defined_for is not None:
            mask = (
                self.calculate(
                    variable.defined_for, period, map_to=variable.entity.key
                )
                > 0
            )
            if np.all(~mask):
                array = holder.default_array()
                array = self._cast_formula_result(array, variable)
                holder.put_in_cache(array, period)

        array = None

        # First, try to run a formula
        try:
            self._check_for_cycle(variable.name, period)
            array = self._run_formula(variable, population, period)

            # If no result, use the default value and cache it
            if array is None:
                # Check if the variable has a previously defined value
                known_periods = holder.get_known_periods()
                if (
                    self.tax_benefit_system.auto_carry_over_input_variables
                    and variable.calculate_output is None
                    and len(known_periods) > 0
                ):
                    # Variables with a calculate-output property specify
                    last_known_period = sorted(known_periods)[-1]
                    array = holder.get_array(last_known_period)
                else:
                    array = holder.default_array()

            if variable.defined_for is not None:
                array = np.where(mask, array, np.zeros_like(array))

            array = self._cast_formula_result(array, variable)
            holder.put_in_cache(array, period)

        except SpiralError:
            array = holder.default_array()
        except RecursionError:
            if self.trace:
                self.tracer.print_computation_log()
            else:
                print(
                    "Computation log is only available with the trace=True option."
                )
            raise Exception(
                f"RecursionError while calculating {variable_name} for period {period}. The full computation trace is printed above."
            )

        return array

    def purge_cache_of_invalid_values(self) -> None:
        # We wait for the end of calculate(), signalled by an empty stack, before purging the cache
        if self.tracer.stack:
            return
        for (_name, _period) in self.invalidated_caches:
            holder = self.get_holder(_name)
            holder.delete_arrays(_period)
        self.invalidated_caches = set()

    def calculate_add(
        self, variable_name: str, period: Period = None
    ) -> ArrayLike:
        variable = self.tax_benefit_system.get_variable(
            variable_name, check_existence=True
        )

        if period is not None and not isinstance(period, Period):
            period = periods.period(period)

        # Check that the requested period matches definition_period
        if periods.unit_weight(
            variable.definition_period
        ) > periods.unit_weight(period.unit):
            raise ValueError(
                "Unable to compute variable '{0}' for period {1}: '{0}' can only be computed for {2}-long periods. You can use the DIVIDE option to get an estimate of {0} by dividing the yearly value by 12, or change the requested period to 'period.this_year'.".format(
                    variable.name, period, variable.definition_period
                )
            )

        if variable.definition_period not in [
            periods.DAY,
            periods.MONTH,
            periods.YEAR,
        ]:
            raise ValueError(
                "Unable to sum constant variable '{}' over period {}: only variables defined daily, monthly, or yearly can be summed over time.".format(
                    variable.name, period
                )
            )

        return sum(
            self.calculate(variable_name, sub_period)
            for sub_period in period.get_subperiods(variable.definition_period)
        )

    def calculate_divide(
        self, variable_name: str, period: Period = None
    ) -> ArrayLike:
        variable = self.tax_benefit_system.get_variable(
            variable_name, check_existence=True
        )

        if period is not None and not isinstance(period, Period):
            period = periods.period(period)

        # Check that the requested period matches definition_period
        if variable.definition_period != periods.YEAR:
            raise ValueError(
                "Unable to divide the value of '{}' over time on period {}: only variables defined yearly can be divided over time.".format(
                    variable_name, period
                )
            )

        if period.size != 1:
            raise ValueError(
                "DIVIDE option can only be used for a one-year or a one-month requested period"
            )

        if period.unit == periods.MONTH:
            computation_period = period.this_year
            return (
                self.calculate(variable_name, period=computation_period) / 12.0
            )
        elif period.unit == periods.YEAR:
            return self.calculate(variable_name, period)

        raise ValueError(
            "Unable to divide the value of '{}' to match period {}.".format(
                variable_name, period
            )
        )

    def calculate_output(
        self, variable_name: str, period: Period = None
    ) -> ArrayLike:
        """
        Calculate the value of a variable using the ``calculate_output`` attribute of the variable.
        """

        variable = self.tax_benefit_system.get_variable(
            variable_name, check_existence=True
        )

        if variable.calculate_output is None:
            return self.calculate(variable_name, period)

        return variable.calculate_output(self, variable_name, period)

    def trace_parameters_at_instant(self, formula_period):
        return TracingParameterNodeAtInstant(
            self.tax_benefit_system.get_parameters_at_instant(formula_period),
            self.tracer,
            self.branch_name,
        )

    def _run_formula(
        self, variable: str, population: Population, period: Period
    ) -> ArrayLike:
        """
        Find the ``variable`` formula for the given ``period`` if it exists, and apply it to ``population``.
        """

        formula = variable.get_formula(period)
        if formula is None:
            return None

        if self.trace:
            parameters_at = self.trace_parameters_at_instant
        else:
            parameters_at = self.tax_benefit_system.get_parameters_at_instant

        if formula.__code__.co_argcount == 2:
            array = formula(population, period)
        else:
            array = formula(population, period, parameters_at)

        return array

    def _check_period_consistency(
        self, period: Period, variable: Variable
    ) -> None:
        """
        Check that a period matches the variable definition_period
        """
        if variable.definition_period == periods.ETERNITY:
            return  # For variables which values are constant in time, all periods are accepted

        if (
            variable.definition_period == periods.MONTH
            and period.unit != periods.MONTH
        ):
            raise ValueError(
                "Unable to compute variable '{0}' for period {1}: '{0}' must be computed for a whole month. You can use the ADD option to sum '{0}' over the requested period, or change the requested period to 'period.first_month'.".format(
                    variable.name, period
                )
            )

        if (
            variable.definition_period == periods.YEAR
            and period.unit != periods.YEAR
        ):
            raise ValueError(
                "Unable to compute variable '{0}' for period {1}: '{0}' must be computed for a whole year. You can use the DIVIDE option to get an estimate of {0} by dividing the yearly value by 12, or change the requested period to 'period.this_year'.".format(
                    variable.name, period
                )
            )

        if period.size != 1:
            raise ValueError(
                "Unable to compute variable '{0}' for period {1}: '{0}' must be computed for a whole {2}. You can use the ADD option to sum '{0}' over the requested period.".format(
                    variable.name,
                    period,
                    "month"
                    if variable.definition_period == periods.MONTH
                    else "year",
                )
            )

    def _cast_formula_result(self, value: Any, variable: str) -> ArrayLike:
        if variable.value_type == Enum and not isinstance(value, EnumArray):
            return variable.possible_values.encode(value)

        if not isinstance(value, numpy.ndarray):
            population = self.get_variable_population(variable.name)
            value = population.filled_array(value)

        if value.dtype != variable.dtype:
            return value.astype(variable.dtype)

        return value

    # ----- Handle circular dependencies in a calculation ----- #

    def _check_for_cycle(self, variable: str, period: Period) -> None:
        """
        Raise an exception in the case of a circular definition, where evaluating a variable for
        a given period loops around to evaluating the same variable/period pair. Also guards, as
        a heuristic, against "quasicircles", where the evaluation of a variable at a period involves
        the same variable at a different period.
        """
        # The last frame is the current calculation, so it should be ignored from cycle detection
        previous_periods = [
            frame["period"]
            for frame in self.tracer.stack[:-1]
            if frame["name"] == variable
            and frame["branch_name"] == self.branch_name
        ]
        if period in previous_periods:
            found_last_frame = False
            i = -2
            while not found_last_frame:
                frame = self.tracer.stack[i]
                if (
                    frame["name"] == variable
                    and frame["branch_name"] == self.branch_name
                ):
                    found_last_frame = True
                i -= 1
            raise CycleError(
                f"Circular definition detected on formula {variable}@{period}. The circle is:\n\nNormal computation tree:\n"
                + "\n".join(
                    f"  {frame['name']}@{frame['period']} (branch {frame['branch_name']})"
                    for frame in self.tracer.stack[:i]
                )
                + "\n\nCycle start:\n"
                + "\n".join(
                    f"  >> {frame['name']}@{frame['period']} (branch {frame['branch_name']})"
                    for frame in self.tracer.stack[i:]
                )
            )
        spiral = len(previous_periods) >= self.max_spiral_loops
        if spiral:
            self.invalidate_spiral_variables(variable)
            message = "Quasicircular definition detected on formula {}@{} involving {}".format(
                variable, period, self.tracer.stack
            )
            raise SpiralError(message, variable)

    def invalidate_cache_entry(self, variable: str, period: Period) -> None:
        self.invalidated_caches.add((variable, period))

    def invalidate_spiral_variables(self, variable: str) -> None:
        # Visit the stack, from the bottom (most recent) up; we know that we'll find
        # the variable implicated in the spiral (max_spiral_loops+1) times; we keep the
        # intermediate values computed (to avoid impacting performance) but we mark them
        # for deletion from the cache once the calculation ends.
        count = 0
        for frame in reversed(self.tracer.stack):
            self.invalidate_cache_entry(frame["name"], frame["period"])
            if frame["name"] == variable:
                count += 1
                if count > self.max_spiral_loops:
                    break

    # ----- Methods to access stored values ----- #

    def get_array(self, variable_name: str, period: Period) -> ArrayLike:
        """
        Return the value of ``variable_name`` for ``period``, if this value is alreay in the cache (if it has been set as an input or previously calculated).

        Unlike :meth:`.calculate`, this method *does not* trigger calculations and *does not* use any formula.
        """
        if period is not None and not isinstance(period, Period):
            period = periods.period(period)
        return self.get_holder(variable_name).get_array(period)

    def get_holder(self, variable_name: str) -> Holder:
        """
        Get the :obj:`.Holder` associated with the variable ``variable_name`` for the simulation
        """
        return self.get_variable_population(variable_name).get_holder(
            variable_name
        )

    def get_memory_usage(self, variables: List[str] = None) -> dict:
        """
        Get data about the virtual memory usage of the simulation
        """
        result = dict(total_nb_bytes=0, by_variable={})
        for entity in self.populations.values():
            entity_memory_usage = entity.get_memory_usage(variables=variables)
            result["total_nb_bytes"] += entity_memory_usage["total_nb_bytes"]
            result["by_variable"].update(entity_memory_usage["by_variable"])
        return result

    # ----- Misc ----- #

    def delete_arrays(self, variable: str, period: Period = None) -> None:
        """
        Delete a variable's value for a given period

        :param variable: the variable to be set
        :param period: the period for which the value should be deleted

        Example:

        >>> from policyengine_core.country_template import CountryTaxBenefitSystem
        >>> simulation = Simulation(CountryTaxBenefitSystem())
        >>> simulation.set_input('age', '2018-04', [12, 14])
        >>> simulation.set_input('age', '2018-05', [13, 14])
        >>> simulation.get_array('age', '2018-05')
        array([13, 14], dtype=int32)
        >>> simulation.delete_arrays('age', '2018-05')
        >>> simulation.get_array('age', '2018-04')
        array([12, 14], dtype=int32)
        >>> simulation.get_array('age', '2018-05') is None
        True
        >>> simulation.set_input('age', '2018-05', [13, 14])
        >>> simulation.delete_arrays('age')
        >>> simulation.get_array('age', '2018-04') is None
        True
        >>> simulation.get_array('age', '2018-05') is None
        True
        """
        self.get_holder(variable).delete_arrays(period)

    def get_known_periods(self, variable: str) -> List[Period]:
        """
        Get a list variable's known period, i.e. the periods where a value has been initialized and

        :param variable: the variable to be set

        Example:

        >>> from policyengine_core.country_template import CountryTaxBenefitSystem
        >>> simulation = Simulation(CountryTaxBenefitSystem())
        >>> simulation.set_input('age', '2018-04', [12, 14])
        >>> simulation.set_input('age', '2018-05', [13, 14])
        >>> simulation.get_known_periods('age')
        [Period((u'month', Instant((2018, 5, 1)), 1)), Period((u'month', Instant((2018, 4, 1)), 1))]
        """
        return self.get_holder(variable).get_known_periods()

    def set_input(
        self, variable_name: str, period: Period, value: ArrayLike
    ) -> None:
        """
        Set a variable's value for a given period

        :param variable: the variable to be set
        :param value: the input value for the variable
        :param period: the period for which the value is setted

        Example:
        >>> from policyengine_core.country_template import CountryTaxBenefitSystem
        >>> simulation = Simulation(CountryTaxBenefitSystem())
        >>> simulation.set_input('age', '2018-04', [12, 14])
        >>> simulation.get_array('age', '2018-04')
        array([12, 14], dtype=int32)

        If a ``set_input`` property has been set for the variable, this method may accept inputs for periods not matching the ``definition_period`` of the variable. To read more about this, check the `documentation <https://openfisca.org/doc/coding-the-legislation/35_periods.html#automatically-process-variable-inputs-defined-for-periods-not-matching-the-definitionperiod>`_.
        """
        variable = self.tax_benefit_system.get_variable(
            variable_name, check_existence=True
        )
        period = periods.period(period)
        if (variable.end is not None) and (period.start.date > variable.end):
            return
        self.get_holder(variable_name).set_input(period, value)

    def get_variable_population(self, variable_name: str) -> Population:
        variable = self.tax_benefit_system.get_variable(
            variable_name, check_existence=True
        )
        return self.populations[variable.entity.key]

    def get_population(self, plural: str = None) -> Population:
        return next(
            (
                population
                for population in self.populations.values()
                if population.entity.plural == plural
            ),
            None,
        )

    def get_entity(self, plural: str = None) -> Entity:
        population = self.get_population(plural)
        return population and population.entity

    def describe_entities(self) -> dict:
        return {
            population.entity.plural: population.ids
            for population in self.populations.values()
        }

    def clone(self, debug: bool = False, trace: bool = False) -> "Simulation":
        """
        Copy the simulation just enough to be able to run the copy without modifying the original simulation
        """
        new = commons.empty_clone(self)
        new_dict = new.__dict__

        for key, value in self.__dict__.items():
            if key not in ("debug", "trace", "tracer"):
                new_dict[key] = value

        new.persons = self.persons.clone(new)
        setattr(new, new.persons.entity.key, new.persons)
        new.populations = {new.persons.entity.key: new.persons}

        for entity in self.tax_benefit_system.group_entities:
            population = self.populations[entity.key].clone(new, new.persons)
            new.populations[entity.key] = population
            setattr(
                new, entity.key, population
            )  # create shortcut simulation.household (for instance)

        new.debug = debug
        new.trace = trace

        return new

    def get_branch(self, name: str = "branch") -> "Simulation":
        """Create a clone of this simulation, whose calculations are traced in the original.

        Args:
            name (str, optional): Name of the branch. Defaults to "branch".
            store (bool, optional): Whether to store the branch in the original simulation. Defaults to True.

        Returns:
            Simulation: The cloned simulation.
        """
        if name == self.branch_name:
            return self
        branch = self.clone()
        self.branches[name] = branch
        branch.branch_name = name
        if self.trace:
            branch.trace = True
            branch.tracer = self.tracer
        return branch



import os
import sys
from typing import TYPE_CHECKING, Iterable, Union

import numpy

from policyengine_core import parameters, tools
from policyengine_core.errors import ParameterNotFoundError
from policyengine_core.parameters import helpers

if TYPE_CHECKING:
    from policyengine_core.parameters.parameter_node import ParameterNode

from policyengine_core.parameters.vectorial_parameter_node_at_instant import (
    VectorialParameterNodeAtInstant,
)


class ParameterNodeAtInstant:
    """
    Parameter node of the legislation, at a given instant.
    """

    def __init__(self, name: str, node: "ParameterNode", instant_str: str):
        """
        :param name: Name of the node.
        :param node: Original :any:`ParameterNode` instance.
        :param instant_str: A date in the format `YYYY-MM-DD`.
        """

        # The "technical" attributes are hidden, so that the node children can be easily browsed with auto-completion without pollution
        self._name = name
        self._instant_str = instant_str
        self._children = {}

        for child_name, child in node.children.items():
            child_at_instant = child._get_at_instant(instant_str)
            if child_at_instant is not None:
                self.add_child(child_name, child_at_instant)

    def add_child(
        self, child_name: str, child_at_instant: "ParameterNodeAtInstant"
    ):
        self._children[child_name] = child_at_instant
        setattr(self, child_name, child_at_instant)

    def __getattr__(self, key: str):
        param_name = helpers._compose_name(self._name, item_name=key)
        raise ParameterNotFoundError(param_name, self._instant_str)

    def __getitem__(
        self, key: str
    ) -> Union["ParameterNodeAtInstant", VectorialParameterNodeAtInstant]:
        # If fancy indexing is used, cast to a vectorial node
        if isinstance(key, numpy.ndarray):
            return parameters.VectorialParameterNodeAtInstant.build_from_node(
                self
            )[key]
        return self._children[key]

    def __iter__(self) -> Iterable:
        return iter(self._children)

    def __repr__(self) -> str:
        result = os.linesep.join(
            [
                os.linesep.join(["{}:", "{}"]).format(
                    name, tools.indent(repr(value))
                )
                for name, value in self._children.items()
            ]
        )
        if sys.version_info < (3, 0):
            return result
        return result



import os
import shutil

import numpy
from numpy.typing import ArrayLike

from policyengine_core import periods
from policyengine_core.enums import EnumArray
from policyengine_core.periods import Period


class OnDiskStorage:
    """
    Low-level class responsible for storing and retrieving calculated vectors on disk
    """

    def __init__(
        self,
        storage_dir: str,
        is_eternal: bool = False,
        preserve_storage_dir: bool = False,
    ):
        self._files = {}
        self._enums = {}
        self.is_eternal = is_eternal
        self.preserve_storage_dir = preserve_storage_dir
        self.storage_dir = storage_dir

    def _decode_file(self, file: str) -> ArrayLike:
        enum = self._enums.get(file)
        if enum is not None:
            return EnumArray(numpy.load(file), enum)
        else:
            return numpy.load(file)

    def get(self, period: Period) -> ArrayLike:
        if self.is_eternal:
            period = periods.period(periods.ETERNITY)
        period = periods.period(period)

        values = self._files.get(period)
        if values is None:
            return None
        return self._decode_file(values)

    def put(self, value: ArrayLike, period: Period) -> None:
        if self.is_eternal:
            period = periods.period(periods.ETERNITY)
        period = periods.period(period)

        filename = str(period)
        path = os.path.join(self.storage_dir, filename) + ".npy"
        if isinstance(value, EnumArray):
            self._enums[path] = value.possible_values
            value = value.view(numpy.ndarray)
        numpy.save(path, value)
        self._files[period] = path

    def delete(self, period: Period = None) -> None:
        if period is None:
            self._files = {}
            return

        if self.is_eternal:
            period = periods.period(periods.ETERNITY)
        period = periods.period(period)

        if period is not None:
            self._files = {
                period_item: value
                for period_item, value in self._files.items()
                if not period.contains(period_item)
            }

    def get_known_periods(self) -> list:
        return list(self._files.keys())

    def restore(self) -> None:
        self._files = files = {}
        # Restore self._files from content of storage_dir.
        for filename in os.listdir(self.storage_dir):
            if not filename.endswith(".npy"):
                continue
            path = os.path.join(self.storage_dir, filename)
            filename_core = filename.rsplit(".", 1)[0]
            period = periods.period(filename_core)
            files[period] = path

    def __del__(self) -> None:
        if self.preserve_storage_dir:
            return
        shutil.rmtree(self.storage_dir)  # Remove the holder temporary files
        # If the simulation temporary directory is empty, remove it
        parent_dir = os.path.abspath(os.path.join(self.storage_dir, os.pardir))
        if not os.listdir(parent_dir):
            shutil.rmtree(parent_dir)



from policyengine_core.projectors import (
    EntityToPersonProjector,
    FirstPersonToEntityProjector,
    Projector,
    UniqueRoleToEntityProjector,
)
from policyengine_core.projectors.helpers import (
    get_projector_from_shortcut,
    projectable,
)

from .config import ADD, DIVIDE
from .group_population import GroupPopulation
from .population import Population



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from policyengine_core.populations import Population

from numpy.typing import ArrayLike

from policyengine_core.projectors.projector import Projector


class EntityToPersonProjector(Projector):
    """For instance person.family."""

    def __init__(self, entity: "Population", parent: Projector = None):
        self.reference_entity = entity
        self.parent = parent

    def transform(self, result: ArrayLike) -> ArrayLike:
        return self.reference_entity.project(result)



from datetime import date

from numpy import logical_not as not_
from numpy import maximum as max_
from numpy import minimum as min_
from numpy import round as round_
from numpy import select, where

from policyengine_core.enums import Enum
from policyengine_core.holders import (
    set_input_dispatch_by_period,
    set_input_divide_by_period,
)
from policyengine_core.parameters import (
    Parameter,
    ParameterNode,
    ParameterScale,
    ParameterScaleBracket,
    load_parameter_file,
)
from policyengine_core.periods import DAY, ETERNITY, MONTH, YEAR, period
from policyengine_core.populations import ADD, DIVIDE
from policyengine_core.reforms import Reform
from policyengine_core.simulations import (
    calculate_output_add,
    calculate_output_divide,
)
from policyengine_core.variables import QuantityType, Variable

STOCK = QuantityType.STOCK
FLOW = QuantityType.FLOW

from policyengine_core.commons.formulas import *



from typing import List

from policyengine_core import entities
from policyengine_core.entities.entity import Entity


def build_entity(
    key: str,
    plural: str,
    label: str,
    doc: str = "",
    roles: List[str] = None,
    is_person: bool = False,
    containing_entities: List[str] = (),
) -> Entity:
    if is_person:
        return entities.Entity(key, plural, label, doc)
    else:
        return entities.GroupEntity(
            key,
            plural,
            label,
            doc,
            roles,
            containing_entities=containing_entities,
        )
